/* This file is part of the xFlows library.
 * Copyright (C) 2009 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of xFlows can be downloaded at 
 * <http://www.lbmethod.org/plbows/>
 *
 * The library xFlows is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Orestis Malaspinas
 */

/** \file
 * A collection of dynamics classes (e.g. BGK) with which a Cell object
 * can be instantiated -- generic implementation.
 */
#ifndef ADV_DIFF_TURBULENT_VISCOSITY_DYNAMICS_HH
#define ADV_DIFF_TURBULENT_VISCOSITY_DYNAMICS_HH

#include "core/latticeStatistics.h"
#include "advDiff_TurbulentViscosity_Dynamics.h"
#include "latticeBoltzmann/advectionDiffusionMomentTemplates.h"
#include "latticeBoltzmann/advectionDiffusionDynamicsTemplates.h"


//namespace plb {
   

/* *************** Class AdvDiff_turbulentViscosity_BGKdynamics *************** */

namespace plb {

    
template<typename T, template<typename U> class Descriptor>
int AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::id =
    meta::registerOneParamDynamics<T,Descriptor,ExternalVelocityParticlesDynamics<T,Descriptor> >("AdvectionDiffusion_TurbulentViscosity");    
    
    
/** \param omega_ relaxation parameter, related to the dynamic viscosity
 */
template<typename T, template<typename U> class Descriptor>
AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::AdvDiff_turbulentViscosity_BGKdynamics (T omega_)
    : AdvectionDiffusionDynamics<T,Descriptor>(omega_)
{ }

template<typename T, template<typename U> class Descriptor>
AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>* AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::clone() const 
{
    return new AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::collide (
        Cell<T,Descriptor>& cell, BlockStatistics& statistics ) 
{
    T rhoBar;
    Array<T,Descriptor<T>::d> jEq;
    advectionDiffusionMomentTemplates<T,Descriptor>::get_rhoBar_jEq(cell, rhoBar, jEq);
    
	T uSqr = noCorrBGKadvDiff_TurbulenViscosity_Collision(cell, rhoBar, jEq,this->getOmega());
    
    if (cell.takesStatistics()) {
        gatherStatistics(statistics, rhoBar, uSqr);
    }
}

/** \param j The parameter j is defined as j = j_advDiff = rho_advDiff*u_fluid
 */
template<typename T, template<typename U> class Descriptor>
T AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::computeEquilibrium (
        plint iPop, T rhoBar, Array<T,Descriptor<T>::d> const& j, T jSqr, T thetaBar ) const
{ 
    return advectionDiffusionDynamicsTemplates<T,Descriptor>::
            bgk_ma1_equilibrium(iPop, rhoBar, j);

}

template<typename T, template<typename U> class Descriptor>
T AdvDiff_turbulentViscosity_BGKdynamics<T,Descriptor>::noCorrBGKadvDiff_TurbulenViscosity_Collision (
        Cell<T,Descriptor>& cell, T rhoBar, Array<T,Descriptor<T>::d> const& jEq, T omega)
{ 
	enum {
		storeturbulentViscosityOffset = // value at the previous iteration K
           		Descriptor<T>::ExternalField::turbulentViscosityBeginsAt,
           		
		storeCellStatus = // is reaction active? 
           		Descriptor<T>::ExternalField::cellStatusBeginsAt,
           		
		storeReactionRateValue = // value at the previous iteration K
           		Descriptor<T>::ExternalField::reactionRateValueBeginsAt,
	};

	
    T turbulentViscosity = cell.getExternal(storeturbulentViscosityOffset);
    // well ..  I just give the tau's value et voila'
	 
    T localConcentration = cell.computeDensity();
	 
    // Normal advection-diffusion collision
    T invRho = Descriptor<T>::invRho(rhoBar);
	const T jSqr = VectorTemplateImpl<T,Descriptor<T>::d>::normSqr(jEq);
    for (plint iPop=0; iPop < Descriptor<T>::q; ++iPop) {
        cell[iPop] *= (T)1-turbulentViscosity;
        cell[iPop] += turbulentViscosity * advectionDiffusionDynamicsTemplates<T,Descriptor>::
            bgk_ma1_equilibrium(iPop, rhoBar, jEq);
    }
    
    
    
    if (*cell.getExternal(storeCellStatus) == 1){ // the cell undergoes to reaction processes (either erosion or sedimentation)
	// The storeCellStatus is a mask. If the cell is close to the boundary (solid walls) and
	// satisfy the condition that I impose for either Sedimentation or Erosion, the mask value is equal to 1,
	// otherwise is value is = 0 (fluid) and a normal advection Diffusion scheme is the result of the collision step.
	reactionRate = *cell.getExternal(storeReactionRateValue) ; 
	for (plint iPop=0; iPop < Descriptor<T>::q; ++iPop) 
	{
	    cell[iPop] += Descriptor<T>::t[iPop]*reactionRate ; // reaction rate (calculated in an external DataProcessor at time t 
								// and saved in an externalField of the AdvDiffusion scheme)
// 	    cell[iPop] += Descriptor<T>::t[i]*localConcentration*kValue ; // reaction rate
	}
    }
	
    return jSqr*invRho*invRho;



}




} // namespace plb

#endif
