/* This file is part of the Palabos library.
 * Copyright (C) 2009 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef POISSON_EQUATION_H
#define POISSON_EQUATION_H

#include "core/globalDefs.h"
#include "atomicBlock/blockLattice3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.h"

namespace plb {

template<typename T1, template<class U> class Descriptor, typename T2>
class BoolMaskFromBounceBack3D : public BoxProcessingFunctional3D_LS<T1,Descriptor,T2>
{
public:
    BoolMaskFromBounceBack3D(plint fluidFlag_=0, plint wallFlag_=1);
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor>& lattice, ScalarField3D<T2>& boolMask);
    virtual BoolMaskFromBounceBack3D<T1,Descriptor,T2>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    plint fluidFlag, wallFlag;
};

template<typename T1, template<class U> class Descriptor, typename T2>
BoolMaskFromBounceBack3D<T1,Descriptor,T2>::BoolMaskFromBounceBack3D(plint fluidFlag_, plint wallFlag_)
    : fluidFlag(fluidFlag_),
      wallFlag(wallFlag_)
{ }

template<typename T1, template<class U> class Descriptor, typename T2>
void BoolMaskFromBounceBack3D<T1,Descriptor,T2>::process (
        Box3D domain, BlockLattice3D<T1,Descriptor>& lattice, ScalarField3D<T2>& boolMask)
{
    Dot3D offset = computeRelativeDisplacement(lattice, boolMask);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if( typeid(lattice.get(iX,iY,iZ).getDynamics()) ==
                    typeid(BounceBack<T1,Descriptor> const&) )
                {
                    boolMask.get(iX+offset.x,iY+offset.y,iZ+offset.z) = wallFlag;
                }
                else {
                    boolMask.get(iX+offset.x,iY+offset.y,iZ+offset.z) = fluidFlag;
                }
            }
        }
    }
}

template<typename T1, template<typename U> class Descriptor, typename T2> 
BoolMaskFromBounceBack3D<T1,Descriptor,T2>* BoolMaskFromBounceBack3D<T1,Descriptor,T2>::clone() const
{
    return new BoolMaskFromBounceBack3D<T1,Descriptor,T2>(*this);
}

template<typename T1, template<typename U> class Descriptor, typename T2>
void BoolMaskFromBounceBack3D<T1,Descriptor,T2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::dataStructure;
}


template<typename T>
class DetectBoundaryFunctional3D : public BoxProcessingFunctional3D_S<T>
{
public:
    virtual void process(Box3D domain, ScalarField3D<T>& data);
    virtual DetectBoundaryFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};

template<typename T>
void DetectBoundaryFunctional3D<T>::process (
        Box3D domain, ScalarField3D<T>& data )
{
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                // If current node is a fluid node ...
                if (data.get(iX,iY,iZ) > 0.5) {
                    T sumNeighbors = data.get(iX-1,iY,iZ) + data.get(iX+1,iY,iZ) +
                                     data.get(iX,iY-1,iZ) + data.get(iX,iY+1,iZ) +
                                     data.get(iX,iY,iZ-1) + data.get(iX,iY,iZ+1);
                    // If any of the neighbors is a wall node ...
                    if (sumNeighbors < 0.5) {
                        // Then current node is defined to be a boundary node.
                        data.get(iX,iY,iZ) = 0.;
                    }
                }
            }
        }
    }
}

template<typename T>
DetectBoundaryFunctional3D<T>* DetectBoundaryFunctional3D<T>::clone() const
{
    return new DetectBoundaryFunctional3D<T>(*this);
}

template<typename T>
void DetectBoundaryFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
}

template<typename T>
class DirichletPoissonFunctional3D : public ReductiveBoxProcessingFunctional3D_SS<T,T>
{
public:
    DirichletPoissonFunctional3D(T rhs_, T omega_);
    virtual void process(Box3D domain, ScalarField3D<T>& data, ScalarField3D<T>& mask);
    virtual DirichletPoissonFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    T getMaximumDifference() const;
private:
    static T bulkLaplaceTerm(ScalarField3D<T>& data, plint iX, plint iY, plint iZ);
    static T boundaryLaplaceTerm(ScalarField3D<T>& data, ScalarField3D<T>& mask,
                                 Dot3D const& offset, plint iX, plint iY, plint iZ);
    static bool isWall(ScalarField3D<T>& boolMask, Dot3D const& offset, plint iX, plint iY, plint iZ);
private:
    T rhs, oneMinusOmega, omegaBySix;
    plint maximumDifferenceId;
};

template<typename T>
DirichletPoissonFunctional3D<T>::DirichletPoissonFunctional3D(T rhs_, T omega_)
    : rhs(rhs_),
      oneMinusOmega(1.-omega_),
      omegaBySix(omega_/6.),
      maximumDifferenceId(this->getStatistics().subscribeMax())
{ }

template<typename T>
T DirichletPoissonFunctional3D<T>::bulkLaplaceTerm(ScalarField3D<T>& data, plint iX, plint iY, plint iZ)
{
    PLB_PRECONDITION(iX>0 && iX<data.getNx()-1);
    PLB_PRECONDITION(iY>0 && iY<data.getNy()-1);
    PLB_PRECONDITION(iZ>0 && iZ<data.getNz()-1);

    return data.get(iX+1,iY,iZ) + data.get(iX-1,iY,iZ) +
           data.get(iX,iY+1,iZ) + data.get(iX,iY-1,iZ) +
           data.get(iX,iY,iZ+1) + data.get(iX,iY,iZ-1);

}

template<typename T>
bool DirichletPoissonFunctional3D<T>::isWall(ScalarField3D<T>& boolMask, Dot3D const& offset, plint iX, plint iY, plint iZ)
{
    return boolMask.get(iX+offset.x,iY+offset.y,iZ+offset.z) < -0.5;
}

template<typename T>
T DirichletPoissonFunctional3D<T>::boundaryLaplaceTerm (
        ScalarField3D<T>& data, ScalarField3D<T>& boolMask,
        Dot3D const& offset, plint iX, plint iY, plint iZ)
{
    PLB_PRECONDITION(iX>0 && iX<data.getNx()-1);
    PLB_PRECONDITION(iY>0 && iY<data.getNy()-1);
    PLB_PRECONDITION(iZ>0 && iZ<data.getNz()-1);

    T xm1 = isWall(boolMask, offset, iX-1,iY,iZ) ? -data.get(iX,iY,iZ) : data.get(iX-1,iY,iZ);
    T xp1 = isWall(boolMask, offset, iX+1,iY,iZ) ? -data.get(iX,iY,iZ) : data.get(iX+1,iY,iZ);
    T ym1 = isWall(boolMask, offset, iX,iY-1,iZ) ? -data.get(iX,iY,iZ) : data.get(iX,iY-1,iZ);
    T yp1 = isWall(boolMask, offset, iX,iY+1,iZ) ? -data.get(iX,iY,iZ) : data.get(iX,iY+1,iZ);
    T zm1 = isWall(boolMask, offset, iX,iY,iZ-1) ? -data.get(iX,iY,iZ) : data.get(iX,iY,iZ-1);
    T zp1 = isWall(boolMask, offset, iX,iY,iZ+1) ? -data.get(iX,iY,iZ) : data.get(iX,iY,iZ+1);

    return  xm1 + xp1 + ym1 + yp1 + zm1 + zp1;
}

template<typename T>
void DirichletPoissonFunctional3D<T>::process (
        Box3D domain, ScalarField3D<T>& data, ScalarField3D<T>& mask )
{
    Dot3D offset = computeRelativeDisplacement(data, mask);
    BlockStatistics& statistics = this->getStatistics();
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                T laplaceTerm = T();
                // Fluid node
                if (mask.get(iX+offset.x,iY+offset.y,iZ+offset.z)>-0.5) {
                    // Non-boundary node
                    if (mask.get(iX+offset.x,iY+offset.y,iZ+offset.z)>0.5) {
                        laplaceTerm = bulkLaplaceTerm(data, iX, iY, iZ);
                    }
                    // Boundary node
                    else {
                        laplaceTerm = boundaryLaplaceTerm(data, mask, offset, iX, iY, iZ);
                    }
                    T newValue = oneMinusOmega*data.get(iX,iY,iZ) + omegaBySix*(laplaceTerm-rhs);
                    T difference = fabs(newValue-data.get(iX,iY,iZ));
                    statistics.gatherMax(maximumDifferenceId, difference);
                    data.get(iX,iY,iZ) = newValue;
                }
            }
        }
    }
}

template<typename T>
DirichletPoissonFunctional3D<T>* DirichletPoissonFunctional3D<T>::clone() const
{
    return new DirichletPoissonFunctional3D<T>(*this);
}

template<typename T>
void DirichletPoissonFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
    modified[1] = modif::nothing;
}

template<typename T>
T DirichletPoissonFunctional3D<T>::getMaximumDifference() const {
    return this->getStatistics().getMax(maximumDifferenceId);
}


template<typename T>
class PoissonToDistanceFunctional3D : public BoxProcessingFunctional3D_S<T>
{
public:
    virtual void process(Box3D domain, ScalarField3D<T>& data);
    virtual PoissonToDistanceFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};

template<typename T>
void PoissonToDistanceFunctional3D<T>::process (
        Box3D domain, ScalarField3D<T>& data )
{
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                data.get(iX,iY,iZ) = 1. - sqrt(1.-fabs(data.get(iX,iY,iZ)));
            }
        }
    }
}

template<typename T>
PoissonToDistanceFunctional3D<T>* PoissonToDistanceFunctional3D<T>::clone() const
{
    return new PoissonToDistanceFunctional3D<T>(*this);
}

template<typename T>
void PoissonToDistanceFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
}

}  // namespace plb

#endif  // POISSON_EQUATION_H
