/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "core/globalDefs.h"
#include "offLattice/voxelizer.h"
#include "offLattice/voxelizer.hh"
#include "plbWrapper/block/dataAnalysisWrapper3D.h"
#include "plbWrapper/block/dataAnalysisWrapper3D.hh"
#include "boundary/meshToVoxel.h"
#include "boundary/meshToVoxel.hh"
#include "atomicBlock/dataField3D.h"
#include "multiBlock/multiBlockGenerator3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.hh"
#include "multiBlock/reductiveMultiDataProcessorWrapper3D.h"
#include "dataProcessors/dataInitializerWrapper3D.h"
#include "dataProcessors/dataInitializerWrapper3D.hh"
#include "dataProcessors/dataAnalysisWrapper3D.h"
#include "dataProcessors/dataAnalysisWrapper3D.hh"
#include "dataProcessors/dataAnalysisFunctional3D.h"
#include "dataProcessors/dataAnalysisFunctional3D.hh"
#include "dataProcessors/metaStuffWrapper3D.h"

namespace plb {

/* ***************** Another Robust Voxelizer ********************************************** */

MultiNTensorField3D<int>* voxelizeHull (
        MultiNTensorField3D<int>& tags, int surfaceTag,
        plint borderWidth )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as outside cells.
    plint envelopeWidth = 1;
    plint nDim = 1;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>((MultiBlock3D&) tags, envelopeWidth, nDim);
    setToConstant( *voxelMatrix, voxelMatrix->getBoundingBox(), voxelFlag::outside );
    setToConstant( *voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), voxelFlag::undetermined );

    plint sumVoxel;
    computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &sumVoxel, nDim);
    bool done = false;
    plint maxIter = 5000;
    plint i=0;
    while (!done && i<maxIter) {
        applyProcessingFunctional (
                new HullVoxelizer3D(surfaceTag),
                voxelMatrix->getBoundingBox(), *voxelMatrix, tags );
        plint nextSumVoxel;
        computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &nextSumVoxel, nDim);
        if (nextSumVoxel == sumVoxel) {
            done = true;
        }
        sumVoxel = nextSumVoxel;
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    setUndeterminedToFlag(*voxelMatrix, voxelFlag::inside);
    //setOneLayerToUndetermined(*voxelMatrix, voxelFlag::inside, voxelFlag::outside);
    //setUndeterminedToFlag(*voxelMatrix, voxelFlag::outside);

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}

MultiNTensorField3D<int>* voxelizeHull (
        MultiNTensorField3D<int>& tags, int surfaceTag,
        plint borderWidth, Box3D seed )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as ouside cells.
    plint envelopeWidth=1;
    plint nDim = 1;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>((MultiBlock3D&) tags, envelopeWidth, nDim);
    setToConstant(*voxelMatrix, voxelMatrix->getBoundingBox(), voxelFlag::undetermined);
    setToConstant(*voxelMatrix, seed, voxelFlag::outside);

    plint sumVoxel;
    computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &sumVoxel, nDim);
    bool done = false;
    plint maxIter = 5000;
    plint i=0;
    while (!done && i<maxIter) {
        applyProcessingFunctional (
                new HullVoxelizer3D(surfaceTag),
                voxelMatrix->getBoundingBox(), *voxelMatrix, tags );
        plint nextSumVoxel;
        computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &nextSumVoxel, nDim);
        if (nextSumVoxel == sumVoxel) {
            done = true;
        }
        sumVoxel = nextSumVoxel;
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    setUndeterminedToFlag(*voxelMatrix, voxelFlag::inside);
    //setOneLayerToUndetermined(*voxelMatrix, voxelFlag::inside, voxelFlag::outside);
    //setUndeterminedToFlag(*voxelMatrix, voxelFlag::outside);

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}


/* ******** HullVoxelizer3D ************************************* */

HullVoxelizer3D::HullVoxelizer3D (
        int surfaceTag_)
    : surfaceTag(surfaceTag_)
{ }

bool HullVoxelizer3D::createVoxelizationRange (
        Box3D const& domain, NTensorField3D<int>& voxels,
        Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange )
{
    // The purpose of the first three loops is to locate the eight
    //   corners of the cube. One voxel per corner would be insufficient
    //   because a potential seed is situated differently, depending on
    //   whether it is on the boundary of the multi-block or somewhere inside.
    for (plint dx=0; dx<=+1; ++dx) {
        plint xMin = domain.x0+dx*domain.getNx()-1;
        plint xMax = domain.x0+dx*domain.getNx();
        for (plint dy=0; dy<=+1; ++dy) {
            plint yMin = domain.y0+dy*domain.getNy()-1;
            plint yMax = domain.y0+dy*domain.getNy();
            for (plint dz=0; dz<=+1; ++dz) {
                plint zMin = domain.z0+dz*domain.getNz()-1;
                plint zMax = domain.z0+dz*domain.getNz();

                // Locate a potential seed in one of the corners.
                for (plint iX=xMin; iX<=xMax; ++iX) {
                    for (plint iY=yMin; iY<=yMax; ++iY) {
                        for (plint iZ=zMin; iZ<=zMax; ++iZ) {
                            if (*voxels.get(iX,iY,iZ) != voxelFlag::undetermined) {
                                xRange[0] = domain.x0+dx*(domain.getNx()-1);
                                xRange[1] = domain.x0+(1-dx)*(domain.getNx()-1);
                                yRange[0] = domain.y0+dy*(domain.getNy()-1);
                                yRange[1] = domain.y0+(1-dy)*(domain.getNy()-1);
                                zRange[0] = domain.z0+dz*(domain.getNz()-1);
                                zRange[1] = domain.z0+(1-dz)*(domain.getNz()-1);
                                return true;
                            }
                        }
                    }
                }

            }
        }
    }
    return false;
}

bool HullVoxelizer3D::voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels,
        NTensorField3D<int> const& tags,
        Dot3D pos, Dot3D neighbor, Dot3D ofs, int& voxelType )
{
    int verificationLevel = 0;
    int typeOfNeighbor = *voxels.get(neighbor.x,neighbor.y,neighbor.z);
    if (typeOfNeighbor==voxelFlag::undetermined) {
        return true;
    }
    // If there is no verification and the voxel has already been voxelized,
    //   it is not being re-voxelized here.
    if (verificationLevel==0) {
        if (voxelType!=voxelFlag::undetermined) {
            return true;
        }
    }
    int newVoxelType = voxelFlag::undetermined;
    int neighborTag = *tags.get(neighbor.x + ofs.x, neighbor.y + ofs.y, neighbor.z + ofs.z);
    if (neighborTag == surfaceTag) {
        return true;
    }
    else {
        newVoxelType = typeOfNeighbor;
    }
    int oldVoxelType = voxelType;
    voxelType = newVoxelType;
    if (oldVoxelType == voxelFlag::undetermined) {
        return true;
    }
    else {
        return oldVoxelType == newVoxelType;
    }
}

void HullVoxelizer3D::process (
        Box3D domain, NTensorField3D<int>& voxels, NTensorField3D<int>& tags )
{
    PLB_PRECONDITION(voxels.getNdim() == tags.getNdim());

    Dot3D ofs = computeRelativeDisplacement(voxels, tags);

    Array<plint,2> xRange, yRange, zRange;
    if (!createVoxelizationRange(domain, voxels, xRange, yRange, zRange)) {
        // If no seed has been found in the envelope, just return and wait
        //   for the next round.
        return;
    }

    // Specify if the loops go in positive or negative direction.
    plint xIncr = xRange[1]>xRange[0] ? 1 : -1;
    plint yIncr = yRange[1]>yRange[0] ? 1 : -1;
    plint zIncr = zRange[1]>zRange[0] ? 1 : -1;
    // The ranges are closed on both ends. Here, the range[1] end
    //   is converted to an open one so we can use != checks in the loops.
    xRange[1] += xIncr; 
    yRange[1] += yIncr; 
    zRange[1] += zIncr; 
    for (plint iX=xRange[0]; iX!=xRange[1]; iX+=xIncr) {
        for (plint iY=yRange[0]; iY!=yRange[1]; iY+=yIncr) {
            for (plint iZ=zRange[0]; iZ!=zRange[1]; iZ+=zIncr) {
                if (*tags.get(iX+ofs.x, iY+ofs.y, iZ+ofs.z) != surfaceTag) {
                    Dot3D pos(iX,iY,iZ);
                    int voxelType = *voxels.get(iX,iY,iZ);
                    if (voxelType==voxelFlag::undetermined) {
                        for (plint dx=-1; dx<=+1; ++dx) {
                            for (plint dy=-1; dy<=+1; ++dy) {
                                for (plint dz=-1; dz<=+1; ++dz) {
                                    if (!(dx==0 && dy==0 && dz==0)) {
                                        Dot3D neighbor(iX+dx, iY+dy, iZ+dz);
                                        bool ok = voxelizeFromNeighbor (
                                                      voxels, tags, 
                                                      pos, neighbor, ofs, voxelType );
                                        PLB_ASSERT( ok );
                                        if (voxelType != voxelFlag::undetermined) {
                                            goto out;
                                        }
                                    }
                                }
                            }
                        }
out:
                        *voxels.get(iX,iY,iZ) = voxelType;
                    }
                }
            }
        }
    }
}

HullVoxelizer3D* HullVoxelizer3D::clone() const {
    return new HullVoxelizer3D(*this);
}

void HullVoxelizer3D::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;   // Voxels
    modified[1] = modif::nothing;           // Node tags
}

BlockDomain::DomainT HullVoxelizer3D::appliesTo() const {
    return BlockDomain::bulk;
}


/* ******** SetUndeterminedToFlagFunctional3D ************************************* */

void setUndeterminedToFlag(MultiNTensorField3D<int>& voxels, int flag)
{
    applyProcessingFunctional(
            new SetUndeterminedToFlagFunctional3D(flag), voxels.getBoundingBox(), voxels);
}

SetUndeterminedToFlagFunctional3D::SetUndeterminedToFlagFunctional3D(int flag_)
    : flag(flag_)
{ }

SetUndeterminedToFlagFunctional3D* SetUndeterminedToFlagFunctional3D::clone() const
{
    return new SetUndeterminedToFlagFunctional3D(*this);
}

void SetUndeterminedToFlagFunctional3D::getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::staticVariables;
}

void SetUndeterminedToFlagFunctional3D::process(Box3D domain, NTensorField3D<int>& voxels)
{
    PLB_PRECONDITION(voxels.getNdim() == 1);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if (*voxels.get(iX,iY,iZ)==voxelFlag::undetermined) {
                    *voxels.get(iX,iY,iZ) = flag;
                }
            }
        }
    }
}


/* ******** SetOneLayerToUndeterminedFunctional3D ************************************* */

void setOneLayerToUndetermined(MultiNTensorField3D<int>& voxels, int flagToReplace, int adjacentFlag)
{
    applyProcessingFunctional(
            new SetOneLayerToUndeterminedFunctional3D(flagToReplace, adjacentFlag),
            voxels.getBoundingBox(), voxels);
}

SetOneLayerToUndeterminedFunctional3D::SetOneLayerToUndeterminedFunctional3D(
        int flagToReplace_, int adjacentFlag_)
    : flagToReplace(flagToReplace_),
      adjacentFlag(adjacentFlag_)
{ }

SetOneLayerToUndeterminedFunctional3D* SetOneLayerToUndeterminedFunctional3D::clone() const
{
    return new SetOneLayerToUndeterminedFunctional3D(*this);
}

void SetOneLayerToUndeterminedFunctional3D::getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::staticVariables;
}

void SetOneLayerToUndeterminedFunctional3D::process(Box3D domain, NTensorField3D<int>& voxels)
{
    PLB_PRECONDITION(voxels.getNdim() == 1);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if (*voxels.get(iX,iY,iZ)==flagToReplace) {
                    for (plint dx=-1; dx<=1; ++dx) {
                        plint nextX = iX + dx;
                        for (plint dy=-1; dy<=1; ++dy) {
                            plint nextY = iY + dy;
                            for (plint dz=-1; dz<=1; ++dz) {
                                plint nextZ = iZ + dz;
                                if (!(dx==0 && dy==0 && dz==0)) {
                                    if (*voxels.get(nextX,nextY,nextZ)==adjacentFlag) {
                                        *voxels.get(iX,iY,iZ) = voxelFlag::undetermined;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}

} // namespace plb

