/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESH_TO_VOXEL_H
#define MESH_TO_VOXEL_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "offLattice/triangleHash.h"
#include "offLattice/voxelizer.h"
#include "geometry/connectedTriangleMesh.h"

namespace plb {

/// This version creates the bounding-box automatically
/// from the mesh.
template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        plint symmetricLayer, plint borderWidth );

template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth );

template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth, Box3D seed );

template<typename T>
void reMeshToVoxel (
        ConnectedTriangleMesh<T>& mesh, MultiNTensorField3D<int>& voxels, plint borderWidth );

/// This data processor accepts one multi-block argument (the
/// voxel matrix) and one class argument (the mesh). The voxel-
/// matrix must be initialized with voxelFlag::undetermined,
/// except for the seed voxels, which are voxelFlag::inside or
/// voxelFlag::outside, and which will not change during
/// voxelization. After successful voxelization, all flags are
/// voxelFlag::inside or voxelFlag::outside.
template<typename T>
class MeshToVoxelFunctional3D : public BoxProcessingFunctional3D {
public:
    MeshToVoxelFunctional3D (
            ConnectedTriangleMesh<T>& mesh_ );
    virtual void processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks );
    virtual MeshToVoxelFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    bool checkIfFacetsCrossed (
            AtomicContainerBlock3D& hashContainer,
            Array<T,3> const& point1, Array<T,3> const& point2,
            T& distance, plint& whichTriangle );
    bool distanceToSurface (
            AtomicContainerBlock3D& hashContainer,
            Array<T,3> const& point, T& distance, bool& isBehind ) const;
    bool createVoxelizationRange (
            Box3D const& domain, NTensorField3D<int>& voxels,
            Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange );
    bool voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels, AtomicContainerBlock3D& hashContainer,
        Dot3D pos, Dot3D neighbor, int& voxelType );
    void printOffender (
            NTensorField3D<int> const& voxels,
            AtomicContainerBlock3D& hashContainer, Dot3D pos );
private:
    ConnectedTriangleMesh<T>& mesh;
};

/// This is a new robust version of the voxelizer algorithm.

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        plint symmetricLayer, plint borderWidth );

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth );

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth, Box3D seed );

/// This data processor is a more robust voxelizer which is
/// expected to work even for badly constructed (non-manifold)
/// surface meshes.
template<typename T>
class Voxelizer3D : public BoxProcessingFunctional3D {
public:
    Voxelizer3D (
            ConnectedTriangleMesh<T>& mesh_ );
    virtual void processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks );
    virtual Voxelizer3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    bool checkIfFacetsCrossed (
            AtomicContainerBlock3D& hashContainer,
            Array<T,3> const& point1, Array<T,3> const& point2,
            T& distance, plint& whichTriangle );
    bool createVoxelizationRange (
            Box3D const& domain, NTensorField3D<int>& voxels,
            Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange );
    bool voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels, AtomicContainerBlock3D& hashContainer,
        Dot3D pos, Dot3D neighbor, int& voxelType );
    void printOffender (
            NTensorField3D<int> const& voxels,
            AtomicContainerBlock3D& hashContainer, Dot3D pos );
private:
    ConnectedTriangleMesh<T>& mesh;
};


/// This is another robust version of the voxelizer algorithm.

MultiNTensorField3D<int>* voxelizeHull (
        MultiNTensorField3D<int>& tags, int surfaceTag,
        plint borderWidth );

MultiNTensorField3D<int>* voxelizeHull (
        MultiNTensorField3D<int>& tags, int surfaceTag,
        plint borderWidth, Box3D seed );

/// This data processor is a more robust voxelizer which is
/// expected to work even for badly constructed (non-manifold)
/// surface meshes.
class HullVoxelizer3D : public BoxProcessingFunctional3D_NN<int,int> {
public:
    HullVoxelizer3D (
            int surfaceTag_ );
    virtual void process (
        Box3D domain, NTensorField3D<int>& voxels, NTensorField3D<int>& tags );
    virtual HullVoxelizer3D* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    bool createVoxelizationRange (
            Box3D const& domain, NTensorField3D<int>& voxels,
            Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange );
    bool voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels,
        NTensorField3D<int> const& tags,
        Dot3D pos, Dot3D neighbor, Dot3D ofs, int& voxelType );
private:
    int surfaceTag;
};


class SetUndeterminedToFlagFunctional3D : public BoxProcessingFunctional3D_N<int> {
public:
    SetUndeterminedToFlagFunctional3D(int flag_);
    virtual SetUndeterminedToFlagFunctional3D* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual void process(Box3D domain, NTensorField3D<int>& voxels);
private:
    int flag;
};

void setUndeterminedToFlag(MultiNTensorField3D<int>& voxels, int flag);


class SetOneLayerToUndeterminedFunctional3D : public BoxProcessingFunctional3D_N<int> {
public:
    SetOneLayerToUndeterminedFunctional3D(int flagToReplace_, int adjacentFlag_);
    virtual SetOneLayerToUndeterminedFunctional3D* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual void process(Box3D domain, NTensorField3D<int>& voxels);
private:
    int flagToReplace;
    int adjacentFlag;
};

void setOneLayerToUndetermined(MultiNTensorField3D<int>& voxels, int flagToReplace, int adjacentFlag);


// Make a copy of the voxel matrix, but with a sparse block-distribution:
// no memory is allocated for blocks which don't have the flag "flag".
MultiNTensorField3D<int>* makeSparse (
        MultiNTensorField3D<int>& fullVoxelMatrix, int flag,
        MultiNTensorField3D<int>& mask, int margin,
        plint blockSize, plint envelopeWidth);

// Same as "makeSparse", except that if blockSize is zero, the distribution
// is not made sparse. Instead, an exact copy is made, with a potentially
// larger envelope width.
MultiNTensorField3D<int>* createSparseVoxelMatrix (
        MultiNTensorField3D<int>& fullVoxelMatrix, int flowType,
        int margin, plint blockSize, plint envelopeWidth);

/// Convert inside flags to innerBoundary, and outside flags to outerBoundary,
///   within a layer of width "borderWidth".
template<typename T>
void identifyBoundary( MultiNTensorField3D<T>& voxelMatrix,
                       Box3D const& domain, plint borderWidth );

template<typename T>
class IdentifyBoundaryFunctional3D : public BoxProcessingFunctional3D_N<T> {
public:
    IdentifyBoundaryFunctional3D(plint borderWidth_);
    virtual void process(Box3D domain, NTensorField3D<T>& voxels);
    virtual IdentifyBoundaryFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    plint borderWidth;
};

template<typename T>
void reIdentifyBoundary( MultiNTensorField3D<T>& voxelMatrix,
                         MultiNTensorField3D<T>& oldVoxelMatrix,
                         Box3D const& domain, plint borderWidth );

template<typename T>
class ReIdentifyBoundaryFunctional3D : public BoxProcessingFunctional3D_NN<T,T> {
public:
    ReIdentifyBoundaryFunctional3D(plint borderWidth_);
    virtual void process(Box3D domain, NTensorField3D<T>& voxels, NTensorField3D<T>& oldVoxels);
    virtual ReIdentifyBoundaryFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    plint borderWidth;
};

template<typename T>
void growDomain( MultiNTensorField3D<T>& matrix,
                 Box3D const& domain, T flag );

template<typename T>
class GrowDomainFunctional3D : public BoxProcessingFunctional3D_N<T> {
public:
    GrowDomainFunctional3D(T flag_);
    virtual void process(Box3D domain, NTensorField3D<T>& voxels);
    virtual GrowDomainFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    T flag;
};

template<typename T>
class BoundaryToUndetFunctional : public BoxProcessingFunctional3D_N<T> {
public:
    virtual void process(Box3D domain, NTensorField3D<T>& voxels);
    virtual BoundaryToUndetFunctional<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
};

template<typename T>
class SparseManagementFunctional3D :
          public PlainReductiveBoxProcessingFunctional3D
{
public:
    SparseManagementFunctional3D(T eliminateFlag_);
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual SparseManagementFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    virtual BlockDomain::DomainT appliesTo() const;
    pluint getNumBlocks() const;
private:
    T eliminateFlag;
    plint numBlocksId;
};

MultiNTensorField3D<int>* fluidMask (
        MultiNTensorField3D<int>& fullVoxelMatrix, int flowType );
    
template<typename T>
MultiBlockManagement3D sparseManagement (
        MultiNTensorField3D<T>& field, T eliminateFlag, plint newEnvelopeWidth );

} // namespace plb

#endif  // MESH_TO_VOXEL_H

