/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SIMULATION_SETUP_3D_HH
#define SIMULATION_SETUP_3D_HH

#include "core/globalDefs.h"
#include "boundary/simulationSetup3D.h"

namespace plb {

// Instantiate the boundary conditions for the outer domain.
template<typename T, template<typename U> class Descriptor>
void outerDomainBoundaries( MultiBlockLattice3D<T,Descriptor>& lattice,
                            MultiNTensorField3D<T>& rhoBar,
                            MultiNTensorField3D<T>& j,
                            OnLatticeBoundaryCondition3D<T,Descriptor>& bc,
                            T inletVelLB, bool lateralFreeSlip )
{
    Array<T,3> uBoundary(inletVelLB, 0.0, 0.0);
    plint nx = lattice.getNx();
    plint ny = lattice.getNy();
    plint nz = lattice.getNz();
    Box3D inlet   (0,      0,      0,      ny-1,   0,      nz-1);
    Box3D outlet  (nx-1,   nx-1,   0,      ny-1,   0,      nz-1);
    Box3D lateral1(1,      nx-2,   0,      0,      0,      nz-1);
    Box3D lateral2(1,      nx-2,   ny-1,   ny-1,   0,      nz-1);
    Box3D lateral3(1,      nx-2,   1,      ny-2,   0,      0);
    Box3D lateral4(1,      nx-2,   1,      ny-2,   nz-1,   nz-1);

    if (lateralFreeSlip) {
        lattice.periodicity().toggleAll(false);
        rhoBar.periodicity().toggleAll(false);
        j.periodicity().toggleAll(false);

        bc.setVelocityConditionOnBlockBoundaries(lattice, inlet, boundary::dirichlet);
        setBoundaryVelocity(lattice, inlet, uBoundary);

        bc.setVelocityConditionOnBlockBoundaries(lattice, lateral1, boundary::freeslip);
        bc.setVelocityConditionOnBlockBoundaries(lattice, lateral2, boundary::freeslip);
        bc.setVelocityConditionOnBlockBoundaries(lattice, lateral3, boundary::freeslip);
        bc.setVelocityConditionOnBlockBoundaries(lattice, lateral4, boundary::freeslip);
        setBoundaryVelocity(lattice, lateral1, uBoundary);
        setBoundaryVelocity(lattice, lateral2, uBoundary);
        setBoundaryVelocity(lattice, lateral3, uBoundary);
        setBoundaryVelocity(lattice, lateral4, uBoundary);

        setBoundaryVelocity(lattice, outlet, uBoundary);
    } else {
        lattice.periodicity().toggleAll(true);
        rhoBar.periodicity().toggleAll(true);
        j.periodicity().toggleAll(true);

        bc.addVelocityBoundary0N(inlet, lattice);
        setBoundaryVelocity(lattice, inlet, uBoundary);
        setBoundaryVelocity(lattice, outlet, uBoundary);
    }
}

template<typename T, template<typename U> class Descriptor, class SurfaceData>
void windTunnelingSetup_along_x (
        Actions3D& lbCycle, Group3D& fields,
        OnLatticeBoundaryCondition3D<T,Descriptor>& outerBoundaryCondition,
        OffLatticeModel3D<T,SurfaceData> const& offLatticeModel ,
        bool lateralFreeSlip, T inletVelLB, T omega, T spongeZoneFraction )
{
    BoxProcessingFunctional3D* collStreamProcessor
        = new ExternalCollideAndStream3D<T,Descriptor>();

    // We use the version of the off-lattice completion functional which does not
    // require rhoBar and j, because otherwise they would need to be recomputed
    // one extra time.
    plint numShapeArgs = 3;
    plint numCompletionArgs = 0;
    /// It is very important that the "offLatticeData" container block
    /// which is passed as the second atomic-block with the off-lattice info
    /// has the same multi-block management as the lattice used in the simulation.
    BoxProcessingFunctional3D* obstacleBCprocessor
        = new OffLatticeCompletionFunctional3D<T,Descriptor,SurfaceData> (
                offLatticeModel.clone(), numShapeArgs, numCompletionArgs );

    // The VirtualOutlet is a sophisticated outflow boundary condition.
    // The "globalDomain" argument for the boundary condition must be
    // bigger than the actual bounding box of the simulation for
    // the directions which are periodic.
    int bcType = 1;
    Box3D globalDomain(fields.getBoundingBox());
    if (!lateralFreeSlip) {
        globalDomain.y0 -= 2; // y-periodicity
        globalDomain.y1 += 2;
        globalDomain.z0 -= 2; // z-periodicity
        globalDomain.z1 += 2;
    }
    T outsideDensity = 1.0;
    BoxProcessingFunctional3D* virtualOutletProcessor
        = new N_VirtualOutlet<T,Descriptor>(outsideDensity, globalDomain, bcType);

    BoxProcessingFunctional3D* rhoBarJprocessor
        = new N_BoxRhoBarJfunctional3D<T,Descriptor>();
                                   
    plint voxelMatrixID   = lbCycle.addBlock(fields.get("voxels"));
    plint hashContainerID = lbCycle.addBlock(fields.get("triangleHash"));
    plint latticeID       = lbCycle.addBlock(fields.get("fluid"));
    plint rhoBarID        = lbCycle.addBlock(fields.get("rhoBar"));
    plint jID             = lbCycle.addBlock(fields.get("j"));
    plint offLattPattID   = lbCycle.addBlock(fields.get("offLatticeData"));
    // Here, the boundary condition needs the lattice as argument (it could also be particles).
    plint boundaryShapeArgID = latticeID;


    lbCycle.addProcessor ( // Originally Level 0
        collStreamProcessor, latticeID, rhoBarID, jID,
        fields.getBoundingBox() );
    lbCycle.addInternalProcessors(latticeID, 0);
    lbCycle.addCommunication(latticeID, modif::staticVariables);

    Box3D outlet(fields.getBoundingBox());
    outlet.x0 = outlet.x1;
    lbCycle.addProcessor (
            virtualOutletProcessor, latticeID, rhoBarID, jID, outlet );

    lbCycle.addProcessor (
            obstacleBCprocessor, latticeID, offLattPattID,       // first two arguments for Guo algorithm.
            voxelMatrixID, hashContainerID, boundaryShapeArgID,  // remaining arguments for inner-flow-shape.
            fields.getBoundingBox() );
    lbCycle.addCommunication(latticeID, modif::staticVariables);

    plint rhoBarJprocID = lbCycle.addProcessor (
        rhoBarJprocessor, latticeID, rhoBarID, jID, fields.getBoundingBox() );
    lbCycle.addCommunication(rhoBarID, modif::staticVariables);
    lbCycle.addCommunication(jID, modif::staticVariables);

    lbCycle.addIncrementTime<T,Descriptor>(latticeID);

    outerDomainBoundaries (
            fields.getLattice<T,Descriptor>("fluid"), fields.getNTensor<T>("rhoBar"),
            fields.getNTensor<T>("j"), outerBoundaryCondition, inletVelLB, lateralFreeSlip );

    // Initial condition: Constant pressure and velocity-at-infinity everywhere.
    Array<T,3> uBoundary(inletVelLB, 0.0, 0.0);
    initializeAtEquilibrium(fields.getLattice<T,Descriptor>("fluid"), fields.getBoundingBox(), 1.0, uBoundary);

    // Compute rhoBar and j before VirtualOutlet is executed.
    lbCycle.execute(rhoBarJprocID, rhoBarJprocID+2);

    plint nx = fields.getBoundingBox().getNx();
    plint ny = fields.getBoundingBox().getNy();
    plint nz = fields.getBoundingBox().getNz();
    plint numOutletSpongeCells = util::roundToInt(spongeZoneFraction*nx);
    if (numOutletSpongeCells > 0) {
        // Number of sponge zone lattice nodes at all the outer domain boundaries.
        // So: 0 means the boundary at x = 0
        //     1 means the boundary at x = nx-1
        //     2 means the boundary at y = 0
        //     and so on...
        Array<plint,6> numSpongeCells;
        numSpongeCells[0] = 0;
        numSpongeCells[1] = numOutletSpongeCells;
        numSpongeCells[2] = 0;
        numSpongeCells[3] = 0;
        numSpongeCells[4] = 0;
        numSpongeCells[5] = 0;

        std::vector<MultiBlock3D*> args;
        args.push_back(&fields.getLattice<T,Descriptor>("fluid"));

        applyProcessingFunctional (
            new ViscositySpongeZone<T,Descriptor> (
                nx, ny, nz, omega, numSpongeCells ),
            fields.getBoundingBox(), args );
    }
}

}  // namespace plb

#endif  // SIMULATION_SETUP_3D_H

