/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TRIANGLE_MESH_SHAPE_HH
#define TRIANGLE_MESH_SHAPE_HH

#include "core/globalDefs.h"
#include "boundary/triangleMeshShape.h"
#include "geometry/triangleHashing.h"
#include "geometry/triangleUtil.h"
#include "offLattice/voxelizer.h"

namespace plb {

template< typename T, class SurfaceData >
TriangleMeshShape3D<T,SurfaceData>::TriangleMeshShape3D (
        ConnectedTriangleMesh<T>& mesh_,
            BoundaryProfileMap<T,SurfaceData> profiles_ )
    : mesh(mesh_),
      bdTagID(mesh.getTriangleTag("boundaryCondition")),
      profiles(profiles_),
      voxelFlags(0), hashContainer(0), boundaryArg(0)
{ }

template< typename T, class SurfaceData >
TriangleMeshShape3D<T,SurfaceData>::TriangleMeshShape3D (
        ConnectedTriangleMesh<T>& mesh_ )
    : mesh(mesh_),
      bdTagID(-1),
      voxelFlags(0), hashContainer(0), boundaryArg(0)
{ }

template< typename T, class SurfaceData >
bool TriangleMeshShape3D<T,SurfaceData>::isInside (
        Dot3D const& location) const
{
    PLB_PRECONDITION( voxelFlags );
    PLB_PRECONDITION( voxelFlags->getNdim()==1 );
    Dot3D localPos = location-voxelFlags->getLocation();
    return voxelFlag::insideFlag(*voxelFlags->get(localPos.x,localPos.y,localPos.z));
}

template< typename T, class SurfaceData >
bool TriangleMeshShape3D<T,SurfaceData>::pointOnSurface (
        Array<T,3> const& fromPoint, Array<T,3> const& direction,
        Array<T,3>& locatedPoint, T& distance,
        Array<T,3>& wallNormal, SurfaceData& surfaceData,
        OffBoundary::Type& bdType, plint& id ) const
{
    PLB_PRECONDITION( hashContainer ); // Make sure these arguments have
    PLB_PRECONDITION( boundaryArg );   //   been provided by the user through
                                       //   the clone function.
    Array<T,3> point2(fromPoint+direction);
    static const T maxDistance = std::sqrt((T)3);
    Array<T,2> xRange(fromPoint[0]-maxDistance, fromPoint[0]+maxDistance);
    Array<T,2> yRange(fromPoint[1]-maxDistance, fromPoint[1]+maxDistance);
    Array<T,2> zRange(fromPoint[2]-maxDistance, fromPoint[2]+maxDistance);
    TriangleHashing<T> triangleHash(*hashContainer);
    std::vector<plint> possibleTriangles;
    if (id>=0 && mesh.triangle(id).get()) {
        possibleTriangles.push_back(id);
    }
    else {
        triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);
    }

    Array<T,3>  tmpLocatedPoint;
    T           tmpDistance;
    Array<T,3>  tmpNormal;
    T shortestDistance = T();
    plint locatedTriangle = -1;

    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint iTriangle = possibleTriangles[iPossible];
        int flag = 0; // Intersection with line segment.
        typename ConnectedTriangleMesh<T>::PTriangle triangle(mesh.triangle(iTriangle));
        bool doesIntersect = pointOnTriangle(fromPoint, point2, flag, *triangle, tmpLocatedPoint, tmpNormal, tmpDistance) == 1;
        if (doesIntersect) 
        {
            if (locatedTriangle==-1 || tmpDistance<shortestDistance) {
                shortestDistance = tmpDistance;
                locatedTriangle = iTriangle;
                locatedPoint = tmpLocatedPoint;
                distance = tmpDistance;
                wallNormal = tmpNormal;
                if (bdTagID >= 0) {
                    plint boundaryConditionTag = mesh.triangle(iTriangle)->tag(bdTagID);
                    profiles[boundaryConditionTag].getData (
                            locatedPoint, iTriangle, boundaryArg, surfaceData, bdType );
                }
            }
        }
    }
    if (locatedTriangle != -1) {
        id = locatedTriangle;
        return true;
    }
    else {
        return false;
    }
}

template<typename T, class SurfaceData>
Array<T,3> TriangleMeshShape3D<T,SurfaceData>::computeContinuousNormal (
        Array<T,3> const& p, plint id, bool isAreaWeighted ) const
{
    return mesh.triangle(id)->continuousNormal(p, isAreaWeighted);
}

template<typename T, class SurfaceData>
bool TriangleMeshShape3D<T,SurfaceData>::intersectsSurface (
        Array<T,3> const& p1, Array<T,3> const& p2, plint& id ) const
{
    PLB_PRECONDITION( hashContainer ); // Make sure these arguments have
    PLB_PRECONDITION( boundaryArg );   //   been provided by the user through
                                       //   the clone function.
    static const T maxDistance = std::sqrt((T)3);
    Array<T,2> xRange(p1[0]-maxDistance, p1[0]+maxDistance);
    Array<T,2> yRange(p1[1]-maxDistance, p1[1]+maxDistance);
    Array<T,2> zRange(p1[2]-maxDistance, p1[2]+maxDistance);
    TriangleHashing<T> triangleHash(*hashContainer);
    std::vector<plint> possibleTriangles;
    if (id>=0 && mesh.triangle(id).get()) {
        possibleTriangles.push_back(id);
    }
    else {
        triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);
    }

    std::vector<plint> selection;
    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint iTriangle = possibleTriangles[iPossible];
        int flag = 0; // Intersection with line segment.
        Array<T,3> intersection, normal;
        T distance;
        typename ConnectedTriangleMesh<T>::PTriangle triangle(mesh.triangle(iTriangle));
        bool doesIntersect = pointOnTriangle(p1, p2, flag, *triangle, intersection, normal, distance) == 1;
        if (doesIntersect) {
            selection.push_back(iTriangle);
        }
    }
    if (selection.empty()) {
        return false;
    }
    else if (selection.size()==1) {
        id = selection[0];
        return true;
    }
    else {
        Array<T,3>  locatedPoint;
        T           distance;
        Array<T,3>  wallNormal;
        SurfaceData surfaceData;  
        OffBoundary::Type bdType;
        return pointOnSurface(p1, p2-p1, locatedPoint, distance, wallNormal, surfaceData, bdType, id);
    }
}

template< typename T, class SurfaceData >
plint TriangleMeshShape3D<T,SurfaceData>::getTag(plint id) const {
    return mesh.triangle(id)->tag(bdTagID);
}

template< typename T, class SurfaceData >
bool TriangleMeshShape3D<T,SurfaceData>::distanceToSurface (
        Array<T,3> const& point, T& distance, bool& isBehind ) const
{
    PLB_PRECONDITION( hashContainer ); // Make sure these arguments have
    PLB_PRECONDITION( boundaryArg );   //   been provided by the user through
                                       //   the clone function.
    T maxDistance = std::sqrt((T)3);
    Array<T,2> xRange(point[0]-maxDistance, point[0]+maxDistance);
    Array<T,2> yRange(point[1]-maxDistance, point[1]+maxDistance);
    Array<T,2> zRange(point[2]-maxDistance, point[2]+maxDistance);
    TriangleHashing<T> triangleHash(*hashContainer);
    std::vector<plint> possibleTriangles;
    triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);

    T    tmpDistance;
    bool tmpIsBehind;
    bool triangleFound = false;

    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint iTriangle = possibleTriangles[iPossible];
        typename ConnectedTriangleMesh<T>::PTriangle triangle(mesh.triangle(iTriangle));
        distanceToTriangle(point, *triangle, tmpDistance, tmpIsBehind);
        if(!triangleFound || tmpDistance<distance) {
            distance = tmpDistance;
            isBehind = tmpIsBehind;
            triangleFound = true;
        }
    }
    return triangleFound;
}

template< typename T, class SurfaceData >
TriangleMeshShape3D<T,SurfaceData>*
    TriangleMeshShape3D<T,SurfaceData>::clone() const
{
    return new TriangleMeshShape3D<T,SurfaceData>(*this);
}

template< typename T, class SurfaceData >
TriangleMeshShape3D<T,SurfaceData>*
    TriangleMeshShape3D<T,SurfaceData>::clone (
            std::vector<AtomicBlock3D*> args ) const
{
    PLB_PRECONDITION( args.size()==3 );
    TriangleMeshShape3D<T,SurfaceData>*
        newShape = new TriangleMeshShape3D<T,SurfaceData>(*this);
    newShape->voxelFlags = dynamic_cast<NTensorField3D<int>*>(args[0]);
    newShape->hashContainer = dynamic_cast<AtomicContainerBlock3D*>(args[1]);
    newShape->boundaryArg = args[2];
    PLB_ASSERT( newShape->voxelFlags );
    PLB_ASSERT( newShape->hashContainer );
    PLB_ASSERT( newShape->boundaryArg );
    return newShape;
}

}  // namespace plb

#endif  // TRIANGLE_MESH_SHAPE_HH

