/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESH_ALGORITHM_HH
#define MESH_ALGORITHM_HH

/*
 * Algorithms, applied to generic meshes.
 */

#include "geometry/meshAlgorithm.h"
#include "core/array.h"
#include <set>
#include <limits>
#include <algorithm>

namespace plb {

template<typename T>    
void toLatticeUnits(TriangleMesh<T>& mesh, T dx) {
    mesh.scale(1.0/dx);
}

template<typename T>    
void toLatticeUnits(TriangleMesh<T>& mesh, Array<T,3> const& center, T dx) {
    Cuboid<T> bCuboid = mesh.getBoundingCuboid();
    Array<T,3> obstacleCenter = (T)0.5 * (bCuboid.lowerLeftCorner + bCuboid.upperRightCorner);
    mesh.translate(-obstacleCenter);
    toLatticeUnits(mesh, dx);
    Array<T,3> centerLB (
            util::roundToInt(center[0]/dx),
            util::roundToInt(center[1]/dx),
            util::roundToInt(center[2]/dx) );
    mesh.translate(centerLB);
}

template<typename T>    
void toPhysicalUnits(TriangleMesh<T>& mesh, T dx, Array<T,3> offset) {
    mesh.scale(dx);
    mesh.translate(offset);
}

template<typename T>    
void toPhysicalUnits(TriangleMesh<T>& mesh, Units3D const& units) {
    toPhysicalUnits(mesh, units.dx(), units.physDomain().lowerLeftCorner);
}

template<typename T>
bool isInteriorVertex(RawConnectedTriangleMesh<T>& triangleMesh, plint vertexID)
{
    std::vector<plint> neighbors;
    findAdjacentVertices(triangleMesh, vertexID, neighbors);
    std::set<T> neighborSet(neighbors.begin(), neighbors.end());
    for (pluint i=0; i<neighbors.size(); ++i) {
        std::vector<plint> neighborNeighbors;
        findAdjacentVertices(triangleMesh, neighbors[i], neighborNeighbors);
        pluint closeNeighbors = 0;
        for (pluint j=0; j<neighborNeighbors.size(); ++j) {
            typename std::set<T>::const_iterator it = neighborSet.find(neighborNeighbors[j]);
            if (it!=neighborSet.end()) {
                ++closeNeighbors;
            }
        }
        if (closeNeighbors<=1) {
            return false;
        }
    }
    return true;
}

template<typename T>
bool isBoundaryVertex(RawConnectedTriangleMesh<T>& triangleMesh, plint vertexID) {
    return !isInteriorVertex(triangleMesh, vertexID);
}

template<typename T>
bool isBoundaryEdge(RawConnectedTriangleMesh<T>& triangleMesh, plint iVertex, plint jVertex) {
    return isBoundaryVertex(triangleMesh, iVertex) && isBoundaryVertex(triangleMesh, jVertex);
}

template<typename T>
void findAdjacentVertices(RawConnectedTriangleMesh<T>& triangleMesh, plint vertexID, std::vector<plint>& neighbors)
{
    typedef typename RawConnectedTriangleMesh<T>::PTriangle PTriangle;
    typedef typename RawConnectedTriangleMesh<T>::PVertex PVertex;
    plint uniqueIDTag = triangleMesh.getVertexTag("UniqueID");
    PVertex vertex = triangleMesh.vertex(vertexID);
    plint numTriangles = vertex->numAdjacentTriangles();
    std::set<plint> neighborSet;
    for (plint iTriangle=0; iTriangle<numTriangles; ++iTriangle) {
        PTriangle triangle = vertex->adjacentTriangle(iTriangle);
        neighborSet.insert(triangle->vertex(0)->tag(uniqueIDTag));
        neighborSet.insert(triangle->vertex(1)->tag(uniqueIDTag));
        neighborSet.insert(triangle->vertex(2)->tag(uniqueIDTag));
    }
    neighbors.assign(neighborSet.begin(), neighborSet.end());
}

template<typename T>
void smooth(RawConnectedTriangleMesh<T>& triangleMesh, plint maxiter, T relax, bool isMeasureWeighted)
{
    typedef typename RawConnectedTriangleMesh<T>::PVertexIterator PVertexIterator;
    typedef typename RawConnectedTriangleMesh<T>::PVertex PVertex;
    plint uniqueIDTag = triangleMesh.getVertexTag("UniqueID");
    T eps0 = std::numeric_limits<T>::epsilon();

    PLB_ASSERT(maxiter >= 0);
    PLB_ASSERT((relax > (T) 0.0 || std::fabs(relax) <= eps0) &&
               (relax < (T) 1.0 || std::fabs(relax - 1.0) <= eps0));

    if (maxiter <= 0)
        return;

    RawConnectedTriangleMesh<T>* bp0;
    bp0 = &triangleMesh;

    RawConnectedTriangleMesh<T> buf(triangleMesh);

    RawConnectedTriangleMesh<T>* bp1;
    bp1 = &buf;

    if (isMeasureWeighted) {
        for (plint iter = 0; iter < maxiter; iter++) {
            PVertexIterator vertexIterator(triangleMesh.vertexIterator());
            while (!vertexIterator->end()) {
                PVertex vertex(vertexIterator->next());
                plint iVertex = vertex->tag(uniqueIDTag);
                Array<T,3> iVertexPos = (*bp0).vertex(iVertex)->get();
                std::vector<plint> neighborVertexIds;
                findAdjacentVertices(*bp0, iVertex, neighborVertexIds);
                if (isInteriorVertex(triangleMesh, iVertex)) {
                    pluint sizeIds = neighborVertexIds.size();
                    T area = (T) 0.0;
                    Array<T,3> tmp((T) 0.0, (T) 0.0, (T) 0.0);
                    for (pluint i = 0; i < sizeIds; i++) {
                        plint jVertex = neighborVertexIds[i];
                        plint kVertex = (i + 1 < sizeIds) ? neighborVertexIds[i + 1] : neighborVertexIds[0];

                        Array<T,3> jVertexPos = (*bp0).vertex(jVertex)->get();
                        Array<T,3> kVertexPos = (*bp0).vertex(kVertex)->get();

                        Array<T,3> middleVertexPos = (iVertexPos + jVertexPos + kVertexPos) / (T) 3.0;

                        Array<T,3> eij = jVertexPos - iVertexPos;
                        Array<T,3> eik = kVertexPos - iVertexPos;
                        Array<T,3> n;
                        crossProduct(eij, eik, n);
                        T locArea = (T) 0.5 * norm(n);

                        area += locArea;
                        tmp += middleVertexPos * locArea;
                    }
                    (*bp1).vertex(iVertex)->get() = ((T) 1.0 - relax) * (*bp0).vertex(iVertex)->get() + relax * tmp / area;
                }
                else
                {  // isInteriorVertex
                    plint jVertex = -1, kVertex = -1;
                    int counter = 0;
                    for (pluint i = 0; i < neighborVertexIds.size(); i++) {
                        plint vertexId = neighborVertexIds[i];
                        if (isBoundaryVertex(triangleMesh, vertexId)) {
                            if (isBoundaryEdge(triangleMesh, iVertex, vertexId)) {
                                counter++;
                                if (counter == 1) {
                                    jVertex = vertexId;
                                } else if (counter == 2) {
                                    kVertex = vertexId;
                                } else {
                                    PLB_ASSERT(false); // Problem with the boundary of the surface mesh.
                                }
                            }
                        }
                    }

                    PLB_ASSERT(counter == 2); // Problem with the boundary of the surface mesh.

                    Array<T,3> jVertexPos = (*bp0).vertex(jVertex)->get();
                    Array<T,3> kVertexPos = (*bp0).vertex(kVertex)->get();

                    Array<T,3> middleVertexPos0 = (iVertexPos + jVertexPos) / (T) 2.0;
                    Array<T,3> middleVertexPos1 = (iVertexPos + kVertexPos) / (T) 2.0;

                    T length0 = norm(jVertexPos - iVertexPos);
                    T length1 = norm(kVertexPos - iVertexPos);

                    (*bp1).vertex(iVertex)->get() = ((T) 1.0 - relax) * (*bp0).vertex(iVertex)->get() +
                                      relax * (middleVertexPos0*length0 + middleVertexPos1*length1) / (length0 + length1);
                }
            }
            // Swap buffers.
            std::swap(bp0, bp1);
        }
    } else {
        for (plint iter = 0; iter < maxiter; iter++) {
            PVertexIterator vertexIterator(triangleMesh.vertexIterator());
            while (!vertexIterator->end()) {
                PVertex vertex(vertexIterator->next());
                plint iVertex = vertex->tag(uniqueIDTag);
                std::vector<plint> neighborVertexIds;
                findAdjacentVertices(triangleMesh, iVertex, neighborVertexIds);

                if (isInteriorVertex(triangleMesh, iVertex)) {
                    pluint sizeIds = neighborVertexIds.size();
                    Array<T,3> tmp((T) 0.0, (T) 0.0, (T) 0.0);
                    for (pluint i = 0; i < sizeIds; i++) {
                        tmp += (*bp0).vertex(neighborVertexIds[i])->get();
                    }
                    (*bp1).vertex(iVertex)->get() = ((T) 1.0 - relax) * (*bp0).vertex(iVertex)->get() + relax * tmp / (T) sizeIds;
                } else {
                    plint jVertex = -1, kVertex = -1;
                    int counter = 0;
                    for (pluint i = 0; i < neighborVertexIds.size(); i++) {
                        plint vertexId = neighborVertexIds[i];
                        if (isBoundaryVertex(triangleMesh, vertexId)) {
                            if (isBoundaryEdge(triangleMesh, iVertex, vertexId)) {
                                counter++;
                                if (counter == 1) {
                                    jVertex = vertexId;
                                } else if (counter == 2) {
                                    kVertex = vertexId;
                                } else {
                                    PLB_ASSERT(false); // Problem with the boundary of the surface mesh.
                                }
                            }
                        }
                    }

                    PLB_ASSERT(counter == 2); // Problem with the boundary of the surface mesh.

                    (*bp1).vertex(iVertex)->get() = ((T) 1.0 - relax) * (*bp0).vertex(iVertex)->get() +
                                      relax * ((*bp0).vertex(jVertex)->get() + (*bp0).vertex(kVertex)->get()) / (T) 2.0;
                }
            }
            // Swap buffers.
            std::swap(bp0, bp1);
        }
    }

    // Final copy of vertex positions.
    if (&triangleMesh != bp0) {
        triangleMesh = *bp0;
    }
}

}  // namespace plb

#endif  // MESH_ALGORITHM_HH

