/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESHES_AND_PARTICLES_HH
#define MESHES_AND_PARTICLES_HH

#include "core/globalDefs.h"
#include "core/util.h"
#include "geometry/meshesAndParticles.h"
#include "particles/visualParticleFunctional3D.h"
#include "particles/particleNonLocalTransfer3D.h"
#include "particles/visualParticle3D.h"
#include "particles/visualParticleWrapper3D.h"
#include "latticeBoltzmann/nearestNeighborLattices3D.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"
#include "multiBlock/multiBlockGenerator3D.h"
#include "dataProcessors/ntensorAnalysisWrapper3D.h"
#include "offLattice/immersedWalls3D.h"

namespace plb {

template<typename T>
MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >*
    createVertexParticles(MultiBlockManagement3D const& management, ConnectedTriangleMesh<T>& mesh)
{
    MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >*
        particleField = new
            MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >
                (management, defaultMultiBlockPolicy3D().getCombinedStatistics());

    Particle3D<T,descriptors::D3Q19Descriptor>* particleTemplate
        = new VisualParticle3D<T,descriptors::D3Q19Descriptor>();

    bool copyVertexProperties = false;
    verticesToParticles(mesh, *particleField, particleTemplate, particleField->getBoundingBox(), copyVertexProperties);

    return particleField;
}

template<typename T>
void scalarFieldToMesh( MultiScalarField3D<T>& scalarField,
                        ConnectedTriangleMesh<T>& mesh, std::string propertyName )
{
    MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >*
        particleField = createVertexParticles(scalarField.getMultiBlockManagement(), mesh);

    std::auto_ptr<MultiScalarField3D<T> > extendedScalarField (
            generateMultiScalarField<T>((MultiBlock3D&)(scalarField), 2) );
    plb::copy(scalarField, *extendedScalarField, scalarField.getBoundingBox());

    scalarFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedScalarField, *particleField, particleField->getBoundingBox() );

    particlesToVertices<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            mesh, *particleField, propertyName, particleField->getBoundingBox() );
    delete particleField;
}

template<typename T>
MultiNTensorField3D<T>* extendEnvelopeWidth(MultiNTensorField3D<T>& field, plint envelopeWidth)
{
    plint nDim = field.getNdim();
    MultiNTensorField3D<T>* extendedField  =
            generateMultiNTensorField3D<T>((MultiBlock3D&)(field), envelopeWidth, nDim);
    plb::copy(field, *extendedField, field.getBoundingBox());
    return extendedField;
}

template<typename T>
void nTensorFieldToMesh( MultiNTensorField3D<T>& nTensorField,
                         ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                         MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    if (!particleField) {
        particleField = createVertexParticles(nTensorField.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(nTensorField, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, *particleField, particleField->getBoundingBox() );
    particlesToVertices<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            mesh, *particleField, propertyName, particleField->getBoundingBox(), nTensorField.getNdim() );
    delete particleField;
    delete extendedNTensorField;
}

template<typename T>
void nTensorFieldToMesh( MultiNTensorField3D<T>& nTensorField,
                         MultiNTensorField3D<int>& mask,
                         ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                         MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    if (!particleField) {
        particleField = createVertexParticles(nTensorField.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(nTensorField, 2);
    MultiNTensorField3D<int>* extendedMask = extendEnvelopeWidth(mask, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, *extendedMask, *particleField, particleField->getBoundingBox() );
    particlesToVertices<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            mesh, *particleField, propertyName, particleField->getBoundingBox(), nTensorField.getNdim() );
    delete particleField;
    delete extendedNTensorField;
    delete extendedMask;
}


template<typename T>
void stressToSurfaceForce( MultiNTensorField3D<T>& stress,
                           ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                           MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    PLB_ASSERT( stress.getNdim()==6 );
    if (!particleField) {
        particleField = createVertexParticles(stress.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(stress, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, particleField, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 6 components of the stress tensor.
    particleStressToForce<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            particleField, mesh, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 3 components of the force. The force can now
    // be integrated over the surface.
    particlesToVertices<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            mesh, particleField, propertyName, particleField->getBoundingBox(), 3 );
    delete particleField;
    delete extendedNTensorField;
}

template<typename T>
void stressToSurfaceForce( MultiNTensorField3D<T>& stress,
                           MultiNTensorField3D<int>& mask,
                           ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                           MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    PLB_ASSERT( stress.getNdim()==6 );
    if (!particleField) {
        particleField = createVertexParticles(stress.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(stress, 2);
    MultiNTensorField3D<int>* extendedMask = extendEnvelopeWidth(mask, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, *extendedMask, *particleField, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 6 components of the stress tensor.
    particleStressToForce<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *particleField, mesh, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 3 components of the force. The force can now
    // be integrated over the surface.
    particlesToVertices<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            mesh, *particleField, propertyName, particleField->getBoundingBox(), 3 );
    delete particleField;
    delete extendedNTensorField;
    delete extendedMask;
}


template<typename T>
Array<T,3> surfaceForceIntegral( MultiNTensorField3D<T>& stress,
                                 MultiNTensorField3D<int>& mask,
                                 ConnectedTriangleMesh<T>& mesh,
                                 MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    PLB_ASSERT( stress.getNdim()==6 );
    if (!particleField) {
        particleField = createVertexParticles(stress.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(stress, 2);
    MultiNTensorField3D<int>* extendedMask = extendEnvelopeWidth(mask, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, *extendedMask, *particleField, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 6 components of the stress tensor.
    particleStressToForce<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *particleField, mesh, particleField->getBoundingBox() );
    std::vector<T> integral = computeSurfaceIntegral< T,descriptors::D3Q19Descriptor,
                                                  DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *particleField, mesh, 3, particleField->getBoundingBox() );
    delete particleField;
    delete extendedNTensorField;
    delete extendedMask;
    PLB_ASSERT( integral.size()==3 );
    return Array<T,3>(integral[0], integral[1], integral[2]);
}


template<typename T>
Array<T,3> surfaceTorqueIntegral( MultiNTensorField3D<T>& stress,
                                  MultiNTensorField3D<int>& mask,
                                  ConnectedTriangleMesh<T>& mesh,
                                  Array<T,3> const& center,
                                  Array<T,3> const& unitaryAxis,
                                  Units3D const& units,
                                  MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField )
{
    PLB_ASSERT( stress.getNdim()==6 );
    if (!particleField) {
        particleField = createVertexParticles(stress.getMultiBlockManagement(), mesh);
    }
    MultiNTensorField3D<T>* extendedNTensorField = extendEnvelopeWidth(stress, 2);
    MultiNTensorField3D<int>* extendedMask = extendEnvelopeWidth(mask, 2);
    nTensorFieldToParticles<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *extendedNTensorField, *extendedMask, *particleField, particleField->getBoundingBox() );
    // Here, the particles (which are parallel), contain the 6 components of the stress tensor.
    particleStressToTorque<T,descriptors::D3Q19Descriptor,DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *particleField, mesh, particleField->getBoundingBox(), center, unitaryAxis );
    std::vector<T> integral = computeSurfaceIntegral< T,descriptors::D3Q19Descriptor,
                                                  DenseParticleField3D<T,descriptors::D3Q19Descriptor> > (
            *particleField, mesh, 3, particleField->getBoundingBox() );
    delete particleField;
    delete extendedNTensorField;
    delete extendedMask;
    PLB_ASSERT( integral.size()==3 );

    // The stress values in the particle field had physical units. Here we need to multiply by dx to go to torque
    // physical units.
    T dx = units.dx();
    return Array<T,3>(dx*integral[0], dx*integral[1], dx*integral[2]);
}

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void verticesToParticles( ConnectedTriangleMesh<T>& mesh,
                          MultiParticleField3D<ParticleFieldT>& particleField,
                          Particle3D<T,Descriptor>* particleTemplate,
                          Box3D domain, bool copyVertexProperties )
{
    std::vector<T> properties;
    plint numProperties = mesh.numVertexProperties();
    if(copyVertexProperties) {
        properties.resize(numProperties);
    }
    plint vertexIDtagging = mesh.getVertexTag("UniqueID");
    std::vector<Particle3D<T,Descriptor>*> particles;
    typename ConnectedTriangleMesh<T>::PVertexIterator it = mesh.vertexIterator();
    while(!it->end()) {
        typename ConnectedTriangleMesh<T>::PVertex vertex = it->next();
        Particle3D<T,Descriptor>* particle = particleTemplate->clone();
        particle->setTag(vertex->tag(vertexIDtagging));
        particle->getPosition() = vertex->get();
        if(copyVertexProperties) {
            for(plint i=0; i<numProperties; ++i) {
                properties[i] = vertex->property(i);
                particle->setScalars(properties);
            }
        }
        particles.push_back(particle);
    }
    delete particleTemplate;

    injectParticlesAtMainProc(particles, particleField, domain);
    for (pluint i=0; i<particles.size(); ++i) {
        delete particles[i];
    }
}

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void surfaceIntegral( ConnectedTriangleMesh<T>& mesh,
                      MultiParticleField3D<ParticleFieldT>& particleField,
                      Box3D domain, plint nDim )
{
    PLB_ASSERT( nDim>0 );
}

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particlesToVertices( ConnectedTriangleMesh<T>& mesh,
                          MultiParticleField3D<ParticleFieldT>& particleField,
                          std::string propertyName, Box3D domain, plint nProperties )
{
    PLB_ASSERT( nProperties>0 );
    std::vector<Particle3D<T,Descriptor>*> particles;
    gatherParticles(particleField, particles, domain);

    if (global::mpi().isMainProcessor()) {
        for (plint iProperty=0; iProperty<nProperties; ++iProperty) {
            std::string iPropertyName;
            if (nProperties==1) {
                iPropertyName = propertyName;
            }
            else {
                iPropertyName = propertyName + "_" + util::val2str(iProperty);
            }
            // Export the values from the particles into a tag->property map.
            std::map<plint,T> properties;
            for(pluint i=0; i<particles.size(); ++i) {
                plint tag = particles[i]->getTag();
                T property;
                particles[i]->getScalar(iProperty, property);
                properties[tag] = property;
            }

            // This function call returns the proper propertyID, even if the property already exists.
            plint propertyId = mesh.registerVertexProperty(iPropertyName);
            plint vertexIDtagging = mesh.getVertexTag("UniqueID");
            PLB_ASSERT( propertyId>=0 );
            typename ConnectedTriangleMesh<T>::PVertexIterator it = mesh.vertexIterator();
            while(!it->end()) {
                typename ConnectedTriangleMesh<T>::PVertex vertex = it->next();
                plint tag = vertex->tag(vertexIDtagging);
                vertex->setProperty(propertyId, properties[tag]);
            }
        }
        for (pluint i=0; i<particles.size(); ++i) {
            delete particles[i];
        }
    }
}



/* ******** ParticleSurfaceIntegral3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
std::vector<T> computeSurfaceIntegral( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh,
                                       plint nDim, Box3D domain )
{
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&particleField);
    ParticleSurfaceIntegral3D<T,Descriptor> dataProcessor(mesh, nDim);
    applyProcessingFunctional(dataProcessor, domain, arg);
    return dataProcessor.getIntegral();
}

template<typename T, template<typename U> class Descriptor>
ParticleSurfaceIntegral3D<T,Descriptor>::ParticleSurfaceIntegral3D(ConnectedTriangleMesh<T>& mesh_, plint nDim_)
    : mesh(&mesh_),
      nDim(nDim_)
{
    for (plint i=0; i<nDim; ++i) {
        integral_ids.push_back(this->getStatistics().subscribeSum());
    }
}

template<typename T, template<typename U> class Descriptor>
void ParticleSurfaceIntegral3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>* particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleField );

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField->findParticles(domain, found);

    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

    std::vector<T> valueIntegral(nDim, T());
    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        plint uniqueID = particle.getTag();
        PVertex vertex = mesh->vertex(uniqueID);
        T area = vertex->area();
        for (plint i=0; i<nDim; ++i) {
            T nextValue;
            particle.getScalar(i, nextValue);
            this->getStatistics().gatherSum(integral_ids[i], nextValue*area);
        }
    }
}


template<typename T, template<typename U> class Descriptor>
ParticleSurfaceIntegral3D<T,Descriptor>*
    ParticleSurfaceIntegral3D<T,Descriptor>::clone() const
{
    return new ParticleSurfaceIntegral3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void ParticleSurfaceIntegral3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing; // Particle field.
}

template<typename T, template<typename U> class Descriptor>
std::vector<T> ParticleSurfaceIntegral3D<T,Descriptor>::getIntegral() const
{
    std::vector<T> integral;
    for (plint i=0; i<nDim; ++i) {
        integral.push_back (
            this->getStatistics().getSum(integral_ids[i]) );
    }
    return integral;
}




/* ******** ParticleStressToForce3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particleStressToForce( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh, Box3D domain)
{
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&particleField);
    applyProcessingFunctional (
            new ParticleStressToForce3D<T,Descriptor>(mesh), domain, arg );
}

template<typename T, template<typename U> class Descriptor>
ParticleStressToForce3D<T,Descriptor>::ParticleStressToForce3D(ConnectedTriangleMesh<T>& mesh_)
    : mesh(&mesh_)
{ }

template<typename T, template<typename U> class Descriptor>
void ParticleStressToForce3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>* particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleField );

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField->findParticles(domain, found);

    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,6> stress;
        for (plint i=0; i<6; ++i) {
            particle.getScalar(i, stress[i]);
        }
        plint uniqueID = particle.getTag();
        PVertex vertex = mesh->vertex(uniqueID);
        Array<T,3> Pi_n;
        SymmetricTensorImpl<T,3>::matVectMult(stress, vertex->normal(), Pi_n);

        std::vector<T> data(3);
        for (plint i=0; i<3; ++i) {
            data[i] = Pi_n[i];
        }
        particle.setScalars(data);
    }
}


template<typename T, template<typename U> class Descriptor>
ParticleStressToForce3D<T,Descriptor>*
    ParticleStressToForce3D<T,Descriptor>::clone() const
{
    return new ParticleStressToForce3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void ParticleStressToForce3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
}




/* ******** ParticleStressToTorque3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particleStressToTorque( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh, Box3D domain,
        Array<T,3> const& center, Array<T,3> const& unitaryAxis)
{
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&particleField);
    applyProcessingFunctional (
            new ParticleStressToTorque3D<T,Descriptor>(mesh, center, unitaryAxis), domain, arg );
}

template<typename T, template<typename U> class Descriptor>
ParticleStressToTorque3D<T,Descriptor>::ParticleStressToTorque3D(ConnectedTriangleMesh<T>& mesh_,
        Array<T,3> const& center_, Array<T,3> const& unitaryAxis_)
    : mesh(&mesh_),
      center(center_),
      unitaryAxis(unitaryAxis_)
{
    T len = norm(unitaryAxis);
    PLB_ASSERT(util::isOne(len));
}

template<typename T, template<typename U> class Descriptor>
void ParticleStressToTorque3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>* particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleField );

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField->findParticles(domain, found);

    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,6> stress;
        for (plint i=0; i<6; ++i) {
            particle.getScalar(i, stress[i]);
        }
        plint uniqueID = particle.getTag();
        PVertex vertex = mesh->vertex(uniqueID);
        Array<T,3> Pi_n;
        SymmetricTensorImpl<T,3>::matVectMult(stress, vertex->normal(), Pi_n);

        Array<T,3> r(vertex->get() - center);
        r -= dot<T,3>(r, unitaryAxis) * unitaryAxis;
        Array<T,3> torque(crossProduct<T>(r, Pi_n));

        std::vector<T> data(3);
        for (plint i=0; i<3; ++i) {
            data[i] = torque[i];
        }
        particle.setScalars(data);
    }
}


template<typename T, template<typename U> class Descriptor>
ParticleStressToTorque3D<T,Descriptor>*
    ParticleStressToTorque3D<T,Descriptor>::clone() const
{
    return new ParticleStressToTorque3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void ParticleStressToTorque3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
}


/* ******** NTensorFieldToParticle3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void nTensorFieldToParticles(MultiNTensorField3D<T>& nTensor, MultiParticleField3D<ParticleFieldT>& particleField, Box3D domain)
{
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&particleField);
    arg.push_back(&nTensor);
    applyProcessingFunctional (
            new NTensorFieldToParticle3D<T,Descriptor>(nTensor.getNdim()), domain, arg );
}

template<typename T, template<typename U> class Descriptor>
NTensorFieldToParticle3D<T,Descriptor>::NTensorFieldToParticle3D(plint nDim_, plint whichScalar_)
    : nDim(nDim_),
      whichScalar(whichScalar_)
{ }

template<typename T, template<typename U> class Descriptor>
void NTensorFieldToParticle3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>* particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleField );
    NTensorField3D<T>* nTensorField =
        dynamic_cast<NTensorField3D<T>*>(blocks[1]);
    PLB_ASSERT( nTensorField );
    PLB_ASSERT( nTensorField->getNdim()==nDim );

    Dot3D partLoc = particleField->getLocation();
    Dot3D nTensorOfs = computeRelativeDisplacement(*particleField, *nTensorField);
    Array<T,3> realLoc((T)partLoc.x,(T)partLoc.y,(T)partLoc.z);

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField->findParticles(domain, found);

    std::vector<Dot3D> cellPos(8);
    std::vector<T> weights(8);
    std::vector<Cell<T,Descriptor>*> cells(8);
    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,3> vertex = particle.getPosition() - realLoc;
        Array<plint,3> intPos (
                (plint)vertex[0], (plint)vertex[1], (plint)vertex[2] );
        std::vector<T> averageData(nDim, T());
        // x   x . x   x
        for (plint dx=-1; dx<=+2; ++dx) {
            for (plint dy=-1; dy<=+2; ++dy) {
                for (plint dz=-1; dz<=+2; ++dz) {
                    Array<plint,3> pos(intPos+Array<plint,3>(dx,dy,dz));
                    Array<T,3> r(pos[0]-vertex[0],pos[1]-vertex[1],pos[2]-vertex[2]);
                    T W = inamuroDeltaFunction<T>().W(r);
                    T* nextData = nTensorField->get(pos[0]+nTensorOfs.x, pos[1]+nTensorOfs.y, pos[2]+nTensorOfs.z);
                    for (plint i=0; i<nDim; ++i) {
                        averageData[i] += W*nextData[i];
                    }
                }
            }
        }

        if (whichScalar==-1) {
            particle.setScalars(averageData);
        }
        else {
            for (plint i=0; i<nDim; ++i) {
                particle.setScalar(whichScalar+i, averageData[i]);
            }
        }
    }
}


template<typename T, template<typename U> class Descriptor>
NTensorFieldToParticle3D<T,Descriptor>*
    NTensorFieldToParticle3D<T,Descriptor>::clone() const
{
    return new NTensorFieldToParticle3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void NTensorFieldToParticle3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // NTensor field.
}


/* ******** MaskedNTensorFieldToParticle3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void nTensorFieldToParticles( MultiNTensorField3D<T>& nTensor, MultiNTensorField3D<int>& mask,
                              MultiParticleField3D<ParticleFieldT>& particleField, Box3D domain)
{
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&particleField);
    arg.push_back(&nTensor);
    arg.push_back(&mask);
    applyProcessingFunctional (
            new MaskedNTensorFieldToParticle3D<T,Descriptor>(nTensor.getNdim()), domain, arg );
}

template<typename T, template<typename U> class Descriptor>
MaskedNTensorFieldToParticle3D<T,Descriptor>::MaskedNTensorFieldToParticle3D(plint nDim_, plint whichScalar_)
    : nDim(nDim_),
      whichScalar(whichScalar_)
{ }

template<typename T, template<typename U> class Descriptor>
void MaskedNTensorFieldToParticle3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==3 );
    ParticleField3D<T,Descriptor>* particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleField );
    NTensorField3D<T>* nTensorField =
        dynamic_cast<NTensorField3D<T>*>(blocks[1]);
    PLB_ASSERT( nTensorField );
    PLB_ASSERT( nTensorField->getNdim()==nDim );
    NTensorField3D<int>* mask =
        dynamic_cast<NTensorField3D<int>*>(blocks[2]);
    PLB_ASSERT( mask );

    Dot3D partLoc = particleField->getLocation();
    Dot3D nTensorOfs = computeRelativeDisplacement(*particleField, *nTensorField);
    Dot3D maskOfs = computeRelativeDisplacement(*particleField, *mask);
    Array<T,3> realLoc((T)partLoc.x,(T)partLoc.y,(T)partLoc.z);

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField->findParticles(domain, found);

    std::vector<Dot3D> cellPos(8);
    std::vector<T> weights(8);
    std::vector<Cell<T,Descriptor>*> cells(8);
    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,3> vertex = particle.getPosition() - realLoc;
        Array<plint,3> intPos (
                (plint)vertex[0], (plint)vertex[1], (plint)vertex[2] );
        std::vector<T> averageData(nDim, T());
        T totalWeight = T();
        // x   x . x   x
        for (plint dx=-1; dx<=+2; ++dx) {
            for (plint dy=-1; dy<=+2; ++dy) {
                for (plint dz=-1; dz<=+2; ++dz) {
                    Array<plint,3> pos(intPos+Array<plint,3>(dx,dy,dz));
                    if (*mask->get(pos[0]+maskOfs.x,pos[1]+maskOfs.y,pos[2]+maskOfs.z)) {
                        Array<T,3> r(pos[0]-vertex[0],pos[1]-vertex[1],pos[2]-vertex[2]);
                        T W = inamuroDeltaFunction<T>().W(r);
                        T* nextData = nTensorField->get(pos[0]+nTensorOfs.x, pos[1]+nTensorOfs.y, pos[2]+nTensorOfs.z);
                        for (plint i=0; i<nDim; ++i) {
                            averageData[i] += W*nextData[i];
                        }
                        totalWeight += W;
                    }
                }
            }
        }
        for (plint i=0; i<nDim; ++i) {
            averageData[i] /= totalWeight;
        }

        if (whichScalar==-1) {
            particle.setScalars(averageData);
        }
        else {
            for (plint i=0; i<nDim; ++i) {
                particle.setScalar(whichScalar+i, averageData[i]);
            }
        }
    }
}


template<typename T, template<typename U> class Descriptor>
MaskedNTensorFieldToParticle3D<T,Descriptor>*
    MaskedNTensorFieldToParticle3D<T,Descriptor>::clone() const
{
    return new MaskedNTensorFieldToParticle3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void MaskedNTensorFieldToParticle3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // NTensor field.
    modified[2] = modif::nothing;  // Mask
}

}  // namespace plb

#endif  // MESHES_AND_PARTICLES_HH

