/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef RAW_TRIANGLE_MESH_HH
#define RAW_TRIANGLE_MESH_HH

#include "core/array.h"
#include "core/globalDefs.h"
#include "io/plbFiles.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"

#include "geometry/rawTriangleMesh.h"
#include "geometry/stlFileIO.h"
#include "geometry/offFileIO.h"

#include <vector>
#include <string>

namespace plb {

/* *************** Free Functions ******************************************* */

template<typename T>
RawTriangleMesh<T> stlToRawTriangleMesh(STLreader<T> const& reader)
{
    return RawTriangleMesh<T>(reader.getParts(), reader.getNames());
}

template<typename T>
RawTriangleMesh<T> stlToRawTriangleMesh(FileName stlFileName, T eps)
{
    return stlToRawTriangleMesh(STLreader<T>(stlFileName.get(), eps));
}

template<typename T>
RawTriangleMesh<T> offToRawTriangleMesh(OFFreader<T> const& reader)
{
    std::vector<Array<T,3> > const& vertices = reader.getVertices();
    std::vector<std::vector<plint> > const& facets = reader.getFacets();
    typedef Array<Array<T,3>,3> RawTriangle;
    std::vector<RawTriangle> triangles(facets.size());

    for (pluint iTriangle = 0; iTriangle < triangles.size(); iTriangle++) {
        PLB_ASSERT(facets[iTriangle].size() == 3); // The surface in the OFF file must be triangulated.
        RawTriangle triangle;
        triangle[0] = vertices[facets[iTriangle][0]];
        triangle[1] = vertices[facets[iTriangle][1]];
        triangle[2] = vertices[facets[iTriangle][2]];
        triangles[iTriangle] = triangle;
    }

    return RawTriangleMesh<T>(triangles);
}

template<typename T>
RawTriangleMesh<T> offToRawTriangleMesh(FileName offFileName)
{
    return offToRawTriangleMesh(OFFreader<T>(offFileName.get()));
}

/* *************** Class RawTriangleMesh::Vertex **************************** */

template<typename T>    
RawTriangleMesh<T>::Vertex::Vertex (
        RawTriangleMesh<T>* mesh_, plint part_, plint triangle_, plint vertex_ )
    : mesh(mesh_),
      part(part_),
      triangle(triangle_),
      vertex(vertex_)
{ }

template<typename T>    
typename RawTriangleMesh<T>::Vertex*
    RawTriangleMesh<T>::Vertex::clone() const
{
    return new typename RawTriangleMesh<T>::Vertex(*this);
}

template<typename T>    
Array<T,3> const& RawTriangleMesh<T>::Vertex::get() const {
    return mesh->triangles[part][triangle][vertex];
}

template<typename T>    
Array<T,3>& RawTriangleMesh<T>::Vertex::get() {
    return mesh->triangles[part][triangle][vertex];
}

template<typename T>    
T const& RawTriangleMesh<T>::Vertex::operator[](plint i) const {
    PLB_ASSERT(i<3);
    return mesh->triangles[part][triangle][vertex][i];
}

template<typename T>    
T& RawTriangleMesh<T>::Vertex::operator[](plint i) {
    PLB_ASSERT(i<3);
    return mesh->triangles[part][triangle][vertex][i];
}

template<typename T>    
Array<T,3> RawTriangleMesh<T>::Vertex::normal(bool areaWeighted) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return Array<T,3>((T) 0, (T) 0, (T) 0);
}

template<typename T>    
T RawTriangleMesh<T>::Vertex::area() const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return ((T) 0);
}

template<typename T>    
plint RawTriangleMesh<T>::Vertex::numAdjacentTriangles() const {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return -1;
}

template<typename T>    
typename RawTriangleMesh<T>::CPTriangle
    RawTriangleMesh<T>::Vertex::adjacentTriangle(plint iTriangle) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return typename RawTriangleMesh<T>::CPTriangle(0);
}

template<typename T>    
typename RawTriangleMesh<T>::PTriangle
    RawTriangleMesh<T>::Vertex::adjacentTriangle(plint iTriangle)
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return typename RawTriangleMesh<T>::PTriangle(0);
}

template<typename T>    
T RawTriangleMesh<T>::Vertex::property(plint whichProperty) const {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return ((T) 0);
}

template<typename T>    
void RawTriangleMesh<T>::Vertex::setProperty(plint whichProperty, T value) {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
}

template<typename T>    
plint RawTriangleMesh<T>::Vertex::tag(plint whichTag) const {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return -1;
}

template<typename T>    
void RawTriangleMesh<T>::Vertex::setTag(plint whichTag, plint value) {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
}


/* *************** Class RawTriangleMesh::Triangle **************************** */
  
  
template<typename T>    
RawTriangleMesh<T>::Triangle::Triangle(RawTriangleMesh<T>* mesh_, plint part_, plint triangle_)
    : mesh(mesh_),
      part(part_),
      triangle(triangle_)
{ }

template<typename T>    
typename RawTriangleMesh<T>::Triangle*
    RawTriangleMesh<T>::Triangle::clone() const
{
    return new typename RawTriangleMesh<T>::Triangle(*this);
}

template<typename T>    
typename RawTriangleMesh<T>::CPVertex
    RawTriangleMesh<T>::Triangle::vertex(plint iVertex) const
{
    PLB_ASSERT(iVertex<3);
    return CPVertex(new Vertex(mesh, part, triangle, iVertex));
}

template<typename T>    
typename RawTriangleMesh<T>::PVertex
    RawTriangleMesh<T>::Triangle::vertex(plint iVertex)
{
    PLB_ASSERT(iVertex<3);
    return PVertex(new Vertex(mesh, part, triangle, iVertex));
}

template<typename T>    
Array<T,3> const& RawTriangleMesh<T>::Triangle::operator[](plint iVertex) const {
    RawTriangle const& data = mesh->triangles[part][triangle];
    return data[iVertex];
}

template<typename T>    
Array<T,3>& RawTriangleMesh<T>::Triangle::operator[](plint iVertex) {
    RawTriangle& data = mesh->triangles[part][triangle];
    return data[iVertex];
}

template<typename T>    
T RawTriangleMesh<T>::Triangle::area() const {
    RawTriangle const& data = mesh->triangles[part][triangle];
    return computeTriangleArea(data[0],data[1],data[2]);
}

template<typename T>    
Array<T,3> RawTriangleMesh<T>::Triangle::normal() const {
    Array<T,3> n(normalTimesArea());
    n /= norm(n);
    return n;
}

template<typename T>    
Array<T,3> RawTriangleMesh<T>::Triangle::normalTimesArea() const {
    RawTriangle const& data = mesh->triangles[part][triangle];
    Array<T,3> e01 = data[1] - data[0];
    Array<T,3> e02 = data[2] - data[0];
    Array<T,3> n;
    crossProduct(e01, e02, n);
    return n;
}

template<typename T>    
Array<T,3> RawTriangleMesh<T>::Triangle::edgeNormal(plint iEdge, bool areaWeighted) const {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return Array<T,3>(T(), T(), T());
}
template<typename T>    
Array<T,3> RawTriangleMesh<T>::Triangle::continuousNormal (
    Array<T,3> const& p, bool areaWeighted ) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return Array<T,3>(T(), T(), T());
}

template<typename T>    
typename RawTriangleMesh<T>::CPTriangle
    RawTriangleMesh<T>::Triangle::edgeNeighbor(plint iEdge) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return CPTriangle(0);
}

template<typename T>    
std::vector<plint> RawTriangleMesh<T>::Triangle::edgeNeighbors(plint iEdge) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return std::vector<plint>();
}

template<typename T>    
plint RawTriangleMesh<T>::Triangle::numVertexNeighbors(plint iVertex) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return 0;
}

template<typename T>    
typename RawTriangleMesh<T>::CPTriangle
    RawTriangleMesh<T>::Triangle::vertexNeighbor(plint iVertex, plint iNeighbor) const
{
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return CPTriangle(0);
}

template<typename T>    
T RawTriangleMesh<T>::Triangle::property(plint whichProperty) const {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
    return ((T) 0.0);
}

template<typename T>    
void RawTriangleMesh<T>::Triangle::setProperty(plint whichProperty, T value) {
    // Not implemented at the level of raw triangle meshes.
    PLB_ASSERT( false );
}

template<typename T>    
plint RawTriangleMesh<T>::Triangle::tag(plint whichTag) const {
    if(whichTag==0) {
        return part;
    }
    else {
        // Not implemented at the level of raw triangle meshes.
        PLB_ASSERT( false );
        return -1;
    }
}

template<typename T>    
void RawTriangleMesh<T>::Triangle::setTag(plint whichTag, plint value) {
    if(whichTag==0) {
        // Can't reassign part numbering.
        PLB_ASSERT( false );
    }
    else {
        // Not implemented at the level of raw triangle meshes.
        PLB_ASSERT( false );
    }
}

/* *************** Class RawTriangleMesh::AllTriangleIterator **************************** */

template<typename T>    
RawTriangleMesh<T>::AllTriangleIterator::AllTriangleIterator(RawTriangleMesh<T>* mesh_)
    : mesh(mesh_),
      currentPart(0),
      currentTriangle(0)
{ }

template<typename T>    
typename RawTriangleMesh<T>::PTriangle RawTriangleMesh<T>::AllTriangleIterator::next()
{
    PLB_ASSERT( currentPart < (plint)mesh->triangles.size() );
    PLB_ASSERT( currentTriangle < (plint)mesh->triangles[currentPart].size() );

    Triangle* triangle = new Triangle(mesh, currentPart, currentTriangle);
    ++currentTriangle;
    if (currentTriangle>=(plint)mesh->triangles[currentPart].size()) {
        currentTriangle=0;
        ++currentPart;
    }
    return PTriangle(triangle);
}

template<typename T>    
bool RawTriangleMesh<T>::AllTriangleIterator::end() const {
    return currentPart >= (plint)mesh->triangles.size();
}


/* *************** Class RawTriangleMesh::PartTriangleIterator **************************** */

template<typename T>    
RawTriangleMesh<T>::PartTriangleIterator::PartTriangleIterator(RawTriangleMesh<T>* mesh_, plint partId_)
    : mesh(mesh_),
      partId(partId_),
      currentTriangle(0)
{ }

template<typename T>    
typename RawTriangleMesh<T>::PTriangle RawTriangleMesh<T>::PartTriangleIterator::next()
{
    PLB_ASSERT( partId < (plint)mesh->triangles.size() );
    PLB_ASSERT( currentTriangle < (plint)mesh->triangles[partId].size() );

    Triangle* triangle = new Triangle(mesh, partId, currentTriangle);
    ++currentTriangle;
    return PTriangle(triangle);
}

template<typename T>    
bool RawTriangleMesh<T>::PartTriangleIterator::end() const {
    return currentTriangle >= (plint)mesh->triangles[partId].size();
}


/* *************** Class RawTriangleMesh  ************************************************* */

template<typename T>    
RawTriangleMesh<T>::RawTriangleMesh(std::vector<typename RawTriangleMesh<T>::RawTriangle> const& triangles_)
{
    // We don't add empty parts, because this would make things more difficult for iterators.
    if (!triangles_.empty()) {
        triangles.push_back(triangles_);
    }
    partNames.push_back("Body");
    computeMinMaxEdges();
    computeBoundingCuboid();
}


template<typename T>    
RawTriangleMesh<T>::RawTriangleMesh(
        std::vector<std::vector<typename RawTriangleMesh<T>::RawTriangle> > const& triangles_,
        std::vector<std::string> const& partNames_ )
{
    PLB_ASSERT( triangles_.size() == partNames_.size() );
    for (pluint i=0; i<triangles_.size(); ++i) {
        // We don't add empty parts, because this would make things more difficult for iterators.
        if (!triangles_[i].empty()) {
            triangles.push_back(triangles_[i]);
            partNames.push_back(partNames_[i]);
        }
    }
    computeMinMaxEdges();
    computeBoundingCuboid();
}

template<typename T>    
RawTriangleMesh<T>::RawTriangleMesh(RawTriangleMesh<T> const& rhs)
    : TriangleMesh<T>(rhs),
      triangles(rhs.triangles),
      partNames(rhs.partNames),
      minEdgeLength(rhs.minEdgeLength),
      maxEdgeLength(rhs.maxEdgeLength),
      boundingCuboid(rhs.boundingCuboid)
{ }

template<typename T>    
RawTriangleMesh<T>::RawTriangleMesh(TriangleMesh<T>& rhs)
{
    for (plint part=0; part<rhs.numParts(); ++part) {
        std::vector<RawTriangle> nextTriangles;
        typename TriangleMesh<T>::PTriangleIterator it = rhs.triangleIterator(part);
        while (!it->end()) {
            typename TriangleMesh<T>::PTriangle triangle = it->next();
            nextTriangles.push_back (
                    RawTriangle( (*triangle)[0],(*triangle)[1],(*triangle)[2]) );
        }
        triangles.push_back(nextTriangles);
        partNames.push_back(rhs.partName(part));
    }
    computeMinMaxEdges();
    computeBoundingCuboid();
}

template<typename T>
RawTriangleMesh<T> RawTriangleMesh<T>::merge(RawTriangleMesh<T> const& rhs) const
{
    std::vector< std::vector<RawTriangle> > newTriangles(triangles);
    newTriangles.insert(newTriangles.end(), rhs.triangles.begin(), rhs.triangles.end());
    std::vector<std::string> newPartNames;
    for (pluint i=0; i<partNames.size(); ++i) {
        newPartNames.push_back("mesh1."+partNames[i]);
    }
    for (pluint i=0; i<rhs.partNames.size(); ++i) {
        newPartNames.push_back("mesh2."+rhs.partNames[i]);
    }
    return RawTriangleMesh(newTriangles, newPartNames);
}

template<typename T>
RawTriangleMesh<T> RawTriangleMesh<T>::refine() const
{
    std::vector<std::vector<RawTriangle> > newTriangles;

    for (pluint part = 0; part < triangles.size(); ++part) {
        std::vector<RawTriangle> nextPart;
        for (pluint i = 0; i < triangles[part].size(); ++i) {
            RawTriangle const& triangle = triangles[part][i];

            Array<T,3> v00 = triangle[0];
            Array<T,3> v11 = triangle[1];
            Array<T,3> v22 = triangle[2];

            Array<T,3> v01 = (T) 0.5 * (v00 + v11);
            Array<T,3> v12 = (T) 0.5 * (v11 + v22);
            Array<T,3> v20 = (T) 0.5 * (v22 + v00);

            RawTriangle newTriangle;

            nextPart.push_back(RawTriangle(v01,v12,v20));
            nextPart.push_back(RawTriangle(v00,v01,v20));
            nextPart.push_back(RawTriangle(v01,v11,v12));
            nextPart.push_back(RawTriangle(v20,v12,v22));
        }
        newTriangles.push_back(nextPart);
    }

    return RawTriangleMesh<T>(newTriangles, partNames);
}


template<typename T>
typename RawTriangleMesh<T>::PTriangleIterator RawTriangleMesh<T>::triangleIterator(plint partId)
{
    if (partId==-1) {
        return PTriangleIterator(new AllTriangleIterator(this));
    }
    else {
        PLB_ASSERT(partId<numParts());
        return PTriangleIterator(new PartTriangleIterator(this, partId));
    }
}

template<typename T>
typename RawTriangleMesh<T>::PVertexIterator RawTriangleMesh<T>::vertexIterator()
{
    // Not implemented for vertices.
    PLB_ASSERT( false );
    return typename RawTriangleMesh<T>::PVertexIterator(0);
}

template<typename T>
plint RawTriangleMesh<T>::numParts() const {
    return (plint) triangles.size();
}

template<typename T>
std::string RawTriangleMesh<T>::partName(plint iPart) const {
    PLB_ASSERT( iPart<(plint)partNames.size() );
    return partNames[iPart];
}

template<typename T>
void RawTriangleMesh<T>::translate(Array<T,3> const& vector)
{
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {
            for (plint j = 0; j < 3; j++) {
                triangles[part][i][j] += vector;
            }
        }
    }
    boundingCuboid.lowerLeftCorner  += vector;
    boundingCuboid.upperRightCorner += vector;
}

template<typename T>
void RawTriangleMesh<T>::scale(T alpha) {
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {
            for (plint j = 0; j < 3; j++) {
                triangles[part][i][j] *= alpha;
            }
        }
    }
    minEdgeLength *= alpha;
    maxEdgeLength *= alpha;

    boundingCuboid.lowerLeftCorner  *= alpha;
    boundingCuboid.upperRightCorner *= alpha;
}

template<typename T>
void RawTriangleMesh<T>::rotateAtOrigin(Array<T,3> const& normedAxis, T theta) {
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {
            for (plint j = 0; j < 3; j++) {
                triangles[part][i][j] =
                    plb::rotateAtOrigin (
                        triangles[part][i][j], normedAxis, theta );
            }
        }
    }

    computeBoundingCuboid();
}

template<typename T>
void RawTriangleMesh<T>::reverseOrientation() {
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {
            std::swap(triangles[part][i][1],triangles[part][i][2]);
        }
    }
}

template<typename T>
T RawTriangleMesh<T>::getMinEdgeLength() const { return minEdgeLength; }

template<typename T>
T RawTriangleMesh<T>::getMaxEdgeLength() const { return maxEdgeLength; }

template<typename T>
Cuboid<T> RawTriangleMesh<T>::getBoundingCuboid() const { return boundingCuboid; }

template<typename T>
std::vector<std::vector<typename RawTriangleMesh<T>::RawTriangle> > const&
    RawTriangleMesh<T>::getTriangles() const
{
    return triangles;
}

template<typename T>
plint RawTriangleMesh<T>::getTriangleTag(std::string tagName) const {
    if(tagName=="Part") {
        return 0;
    }
    return -1;
}
template<typename T>
std::string RawTriangleMesh<T>::getTriangleTagName(plint tag) const {
    PLB_ASSERT(tag==0);
    return "Part";
}

template<typename T>
plint RawTriangleMesh<T>::numTriangleTags() const {
    return 1;
}

template<typename T>
void RawTriangleMesh<T>::computeMinMaxEdges() {
    minEdgeLength = std::numeric_limits<T>::max();
    maxEdgeLength = std::numeric_limits<T>::min();
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {

            RawTriangle const& triangle = triangles[part][i];
            T edge1 = norm(triangle[1]-triangle[0]);
            T edge2 = norm(triangle[2]-triangle[1]);
            T edge3 = norm(triangle[0]-triangle[2]);
            T minEdge = std::min(edge1, std::min(edge2, edge3));
            T maxEdge = std::max(edge1, std::max(edge2, edge3));
            minEdgeLength = std::min(minEdgeLength, minEdge);
            maxEdgeLength = std::max(maxEdgeLength, maxEdge);
        }
    }
}

template<typename T>
void RawTriangleMesh<T>::computeBoundingCuboid() {
    T xMin, yMin, zMin;
    T xMax, yMax, zMax;

    xMin = yMin = zMin =  std::numeric_limits<T>::max();
    xMax = yMax = zMax = -std::numeric_limits<T>::max();
    for (pluint part=0; part<triangles.size(); ++part) {
        for (pluint i=0; i<triangles[part].size(); i++) {
            RawTriangle const& triangle = triangles[part][i];

            xMin = std::min(xMin, std::min(triangle[0][0], std::min(triangle[1][0], triangle[2][0])));
            yMin = std::min(yMin, std::min(triangle[0][1], std::min(triangle[1][1], triangle[2][1])));
            zMin = std::min(zMin, std::min(triangle[0][2], std::min(triangle[1][2], triangle[2][2])));

            xMax = std::max(xMax, std::max(triangle[0][0], std::max(triangle[1][0], triangle[2][0])));
            yMax = std::max(yMax, std::max(triangle[0][1], std::max(triangle[1][1], triangle[2][1])));
            zMax = std::max(zMax, std::max(triangle[0][2], std::max(triangle[1][2], triangle[2][2])));
        }
    }
    boundingCuboid.lowerLeftCorner  = Array<T,3>(xMin, yMin, zMin);
    boundingCuboid.upperRightCorner = Array<T,3>(xMax, yMax, zMax);
}

template<typename T>
plint RawTriangleMesh<T>::getNumTriangles() const
{
    plint numTriangles = 0;
    for (pluint part = 0; part < triangles.size(); ++part) {
        numTriangles += triangles[part].size();
    }
    return numTriangles;
}

}  // namespace plb

#endif  // RAW_TRIANGLE_MESH_HH

