/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2014 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef STREAMLINE_PARTICLE_WRAPPER_3D_HH
#define STREAMLINE_PARTICLE_WRAPPER_3D_HH

#include "core/globalDefs.h"
#include "core/array.h"
#include "core/geometry3D.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.h"
#include "particles/visualParticleWrapper3D.h"
#include "particles/visualParticleFunctional3D.h"
#include "particles/particleProcessingFunctional3D.h"
#include "particles/particleNonLocalTransfer3D.h"
#include "geometry/streamlineParticle3D.h"
#include "geometry/streamlineParticleWrapper3D.h"
#include "geometry/connectedTriangleUtil.h"
#include "geometry/obsoleteFormatWrapper.h"
#include "particles/particleProcessingFunctional3D.h"
#include "latticeBoltzmann/nearestNeighborLattices3D.h"

#include <vector>
#include <cmath>

namespace plb {

template<typename T>
RawConnectedTriangleMesh<T> generateSphereStreamlines (
        MultiTensorField3D<T,3>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T sphereRadius,
        plint sphereMinNumOfTriangles, plint numSteps,
        plint numResamplingPoints )
{
    bool isNtensor = false;
    return generateSphereStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step, sphereRadius,
            sphereMinNumOfTriangles, numSteps, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateSphereStreamlines (
        MultiNTensorField3D<T>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T sphereRadius,
        plint sphereMinNumOfTriangles, plint numSteps,
        plint numResamplingPoints )
{
    bool isNtensor = true;
    return generateSphereStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step, sphereRadius,
            sphereMinNumOfTriangles, numSteps, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateSphereStreamlines_generic (
        MultiBlock3D& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T sphereRadius,
        plint sphereMinNumOfTriangles, plint numSteps,
        plint numResamplingPoints, bool isNtensor )
{
    // When a constant spatial step is used, this must not be greater than one cell.
    if (useConstantSpatialStepIntegration) {
        PLB_ASSERT(std::fabs(step) > getEpsilon<T>(floatingPointPrecision<T>()));
        PLB_ASSERT(std::fabs(step) < (T) 1);
    }

    Box3D domain = velocityField.getBoundingBox();
    SphereStreamlineParticle3D<T,descriptors::D3Q19Descriptor> *particleTemplate =
        new SphereStreamlineParticle3D<T,descriptors::D3Q19Descriptor> (
                0, Array<T,3>((T)0.,(T)0.,(T)0.), Array<T,3>((T)0.,(T)0.,(T)0.),
                useConstantSpatialStepIntegration, step, domain,
                sphereRadius, sphereMinNumOfTriangles );

    return generateStreamlines (
            velocityField, particleTemplate, seedingPositions, numSteps, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateTubeStreamlines (
        MultiTensorField3D<T,3>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T tubeNormalRadius,
        T tubeBinormalRadius, plint tubeNumPoints, plint numSteps,
        bool useFrenetSerret, plint numResamplingPoints )
{
    bool isNtensor = false;
    return generateTubeStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step, tubeNormalRadius,
            tubeBinormalRadius, tubeNumPoints, numSteps, useFrenetSerret, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateTubeStreamlines (
        MultiNTensorField3D<T>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T tubeNormalRadius,
        T tubeBinormalRadius, plint tubeNumPoints, plint numSteps,
        bool useFrenetSerret, plint numResamplingPoints )
{
    bool isNtensor = true;
    return generateTubeStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step, tubeNormalRadius,
            tubeBinormalRadius, tubeNumPoints, numSteps, useFrenetSerret, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T>* swig_generateTubeStreamlines (
        MultiNTensorField3D<T>& velocityField,
        T* seedingPositions, int size,
        bool useConstantSpatialStepIntegration, T step, T tubeNormalRadius,
        T tubeBinormalRadius, plint tubeNumPoints, plint numSteps,
        bool useFrenetSerret, plint numResamplingPoints )
{
    PLB_ASSERT( size%3==0 );
    std::vector<Array<T,3> > positions;
    plint i=0;
    while(i<size) {
        Array<T,3> pos;
        pos[0] = seedingPositions[i]; ++i;
        pos[1] = seedingPositions[i]; ++i;
        pos[2] = seedingPositions[i]; ++i;
        positions.push_back(pos);
    }
    return new RawConnectedTriangleMesh<T> (
            generateTubeStreamlines(velocityField, positions, useConstantSpatialStepIntegration,
                                    step, tubeNormalRadius, tubeBinormalRadius, tubeNumPoints,
                                    numSteps, useFrenetSerret, numResamplingPoints) );
}

template<typename T>
RawConnectedTriangleMesh<T> generateTubeStreamlines_generic (
        MultiBlock3D& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step, T tubeNormalRadius,
        T tubeBinormalRadius, plint tubeNumPoints, plint numSteps,
        bool useFrenetSerret, plint numResamplingPoints, bool isNtensor )
{
    // When a constant spatial step is used, this must not be greater than one cell.
    if (useConstantSpatialStepIntegration) {
        PLB_ASSERT(std::fabs(step) > getEpsilon<T>(floatingPointPrecision<T>()));
        PLB_ASSERT(std::fabs(step) < (T) 1);
    }

    Box3D domain = velocityField.getBoundingBox();
    TubeStreamlineParticle3D<T,descriptors::D3Q19Descriptor> *particleTemplate =
        new TubeStreamlineParticle3D<T,descriptors::D3Q19Descriptor> (
                0, Array<T,3>((T)0.,(T)0.,(T)0.), Array<T,3>((T)0.,(T)0.,(T)0.),
                useConstantSpatialStepIntegration, step, domain,
                tubeNormalRadius, tubeBinormalRadius, tubeNumPoints,
                useFrenetSerret);

    return generateStreamlines (
            velocityField, particleTemplate, seedingPositions, numSteps, numResamplingPoints, isNtensor );
}


template<typename T>
RawConnectedTriangleMesh<T> generateStripStreamlines (
        MultiTensorField3D<T,3>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step,
        T stripThickness, plint numSteps, bool useFrenetSerret,
        plint numResamplingPoints )
{
    bool isNtensor = false;
    return generateStripStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step,
            stripThickness, numSteps, useFrenetSerret, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateStripStreamlines (
        MultiNTensorField3D<T>& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step,
        T stripThickness, plint numSteps, bool useFrenetSerret,
        plint numResamplingPoints )
{
    bool isNtensor = true;
    return generateStripStreamlines_generic (
            velocityField, seedingPositions, useConstantSpatialStepIntegration, step,
            stripThickness, numSteps, useFrenetSerret, numResamplingPoints, isNtensor );
}

template<typename T>
RawConnectedTriangleMesh<T> generateStripStreamlines_generic (
        MultiBlock3D& velocityField,
        std::vector<Array<T,3> > const& seedingPositions,
        bool useConstantSpatialStepIntegration, T step,
        T stripThickness, plint numSteps, bool useFrenetSerret,
        plint numResamplingPoints, bool isNtensor )
{
    // When a constant spatial step is used, this must not be greater than one cell.
    if (useConstantSpatialStepIntegration) {
        PLB_ASSERT(std::fabs(step) > getEpsilon<T>(floatingPointPrecision<T>()));
        PLB_ASSERT(std::fabs(step) < (T) 1);
    }

    Box3D domain = velocityField.getBoundingBox();
    StripStreamlineParticle3D<T,descriptors::D3Q19Descriptor> *particleTemplate =
        new StripStreamlineParticle3D<T,descriptors::D3Q19Descriptor>(
                0, Array<T,3>((T)0.,(T)0.,(T)0.), Array<T,3>((T)0.,(T)0.,(T)0.),
                useConstantSpatialStepIntegration, step, domain,
                stripThickness, useFrenetSerret);

    return generateStreamlines (
            velocityField, particleTemplate, seedingPositions, numSteps, numResamplingPoints, isNtensor );
}

template<typename T, template<typename U> class Descriptor>
RawConnectedTriangleMesh<T> generateStreamlines (
        MultiBlock3D& velocityField,
        Particle3D<T,Descriptor>* particleTemplate,
        std::vector<Array<T,3> > const& seedingPositions,
        plint numSteps, plint numResamplingPoints, bool isNtensor )
{
    pluint size = seedingPositions.size();
    std::vector<Particle3D<T,Descriptor>*> streamlineParticles(size);
    for (pluint iPosition = 0; iPosition < size; iPosition++) {
        Particle3D<T,Descriptor>* nextParticle = particleTemplate->clone();
        nextParticle->setTag(iPosition);
        nextParticle->getPosition() = seedingPositions[iPosition];
        streamlineParticles[iPosition] = nextParticle;
    }
    delete particleTemplate;

    MultiParticleField3D<LightParticleField3D<T,Descriptor> > *streamlineParticleField =
        new MultiParticleField3D<LightParticleField3D<T,Descriptor> > (
                velocityField.getMultiBlockManagement(),
                defaultMultiBlockPolicy3D().getCombinedStatistics() );

    std::vector<MultiBlock3D*> particleVelocityArg;
    particleVelocityArg.push_back(streamlineParticleField);
    particleVelocityArg.push_back(&velocityField);

    T timeFactor = 1.0;
    if (isNtensor) {
        integrateProcessingFunctional(
                new N_VelocityToParticleCoupling3D<T,Descriptor>(timeFactor),
                streamlineParticleField->getBoundingBox(), particleVelocityArg, 0);
    }
    else {
        integrateProcessingFunctional(
                new VelocityToParticleCoupling3D<T,Descriptor>(timeFactor),
                streamlineParticleField->getBoundingBox(), particleVelocityArg, 0);
    }

    std::vector<MultiBlock3D*> particleArg;
    particleArg.push_back(streamlineParticleField);

    T cutOffSpeedSqr = -1.0;
    integrateProcessingFunctional(
            new AdvanceParticlesFunctional3D<T,Descriptor>(cutOffSpeedSqr),
            streamlineParticleField->getBoundingBox(), particleArg, 1);

    // Inject the streamline particles.
    injectParticles<T,Descriptor,LightParticleField3D<T,Descriptor> >(streamlineParticles, *streamlineParticleField,
            streamlineParticleField->getBoundingBox());

    for (pluint iParticle = 0; iParticle < streamlineParticles.size(); iParticle++) {
        delete streamlineParticles[iParticle];
    }
    streamlineParticles.clear();

    // Perform the streamline integration.
    for (plint iStep = 0; iStep < numSteps; iStep++) {
        streamlineParticleField->executeInternalProcessors();
    }

    // Gather all streamline particles to processor 0.
    Box3D domain = streamlineParticleField->getBoundingBox();
    SparseBlockStructure3D blockStructure(domain);
    blockStructure.addBlock(domain, 0);
    plint envelopeWidth=1;
    MultiBlockManagement3D serialMultiBlockManagement(
            blockStructure, new OneToOneThreadAttribution, envelopeWidth);

    MultiParticleField3D<LightParticleField3D<T,Descriptor> > multiSerialParticles(
            serialMultiBlockManagement,
            defaultMultiBlockPolicy3D().getCombinedStatistics());

    copy(*streamlineParticleField, domain, multiSerialParticles, domain);

    TriangleSet<T> streamlines;
    if (global::mpi().isMainProcessor()) {
        ParticleField3D<T,Descriptor>& atomicSerialParticles =
            dynamic_cast<ParticleField3D<T,Descriptor>&>(multiSerialParticles.getComponent(0));

        std::vector<Particle3D<T,Descriptor>*> found;
        SmartBulk3D oneBlockBulk(serialMultiBlockManagement, 0);
        atomicSerialParticles.findParticles(oneBlockBulk.toLocal(domain), found);

        TriangleSet<T> streamline(DBL);
        for (pluint iParticle = 0; iParticle < found.size(); iParticle++) {
            StreamlineParticle3D<T,Descriptor>* streamlineParticle =
                dynamic_cast<StreamlineParticle3D<T,Descriptor>*>(found[iParticle]);
            PLB_ASSERT(streamlineParticle);
            streamlineParticle->resample(numResamplingPoints);
            streamlineParticle->createStreamline(streamline);
            streamlines.append(streamline);
        }
    }
    RawTriangleMesh<T> triangleMesh(triangleSetToRawTriangleMesh(streamlines));
    return MeshConnector<T>(triangleMesh).generateConnectedMesh();
}

}  // namespace plb

#endif  // STREAMLINE_PARTICLE_WRAPPER_3D_HH

