/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VTK_FILE_IO_HH
#define VTK_FILE_IO_HH

#include <limits>
#include <fstream>
#include "geometry/vtkFileIO.h"
#include "io/imageWriter.h"
#include "geometry/connectedTriangleUtil.h"

namespace plb {

template<typename T>
void writeVTK(RawConnectedTriangleMesh<T>& mesh, FileName fname)
{
    //pcout << "parts:" << mesh.numParts() << " " << mesh.getNumTriangles() << std::endl;
    if (!global::mpi().isMainProcessor()) return;
    if (mesh.getNumTriangles() == 0) return;
    typedef typename ConnectedTriangleMesh<T>::PTriangleIterator PTriangleIterator;
    typedef typename ConnectedTriangleMesh<T>::PVertexIterator PVertexIterator;
    typedef typename ConnectedTriangleMesh<T>::PTriangle PTriangle;
    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

   // pcout << "Number of parts in the file: " << mesh.numParts() << std::endl;
    for (plint iPart=0; iPart<mesh.numParts(); ++iPart) {
        FileName partFileName(fname);
        if (mesh.numParts() > 1) {
            partFileName.setName(createFileName(partFileName.getName()+"_", iPart, 3));
        }
        RawConnectedTriangleMesh<T> part = extractConnectedPart(mesh, iPart);
        pcout << "Creating file " << partFileName.get() << std::endl;
        std::ofstream ofile(partFileName.get().c_str());
        ofile.precision(10);
        std::scientific(ofile);

        ofile << "# vtk DataFile Version 3.0\n";
        ofile << "Surface mesh created with Palabos\n";
        ofile << "ASCII\n";
        ofile << "DATASET UNSTRUCTURED_GRID\n";

        ofile << "POINTS " << part.getNumVertices()
              << (sizeof(T)==sizeof(double) ? " double" : " float")
              << "\n";

        plint vertexIDtagging = part.getVertexTag("UniqueID");
        PVertexIterator vertexIt = part.vertexIterator();
        std::map<plint,plint> toNewVertexID;
        plint newVertexID=0;
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            ofile << (*vertex)[0] << " " << (*vertex)[1] << " " << (*vertex)[2] << "\n";
            toNewVertexID[vertex->tag(vertexIDtagging)] = newVertexID;
            ++newVertexID;
        }
        ofile << "\n";

        ofile << "CELLS " << part.getNumTriangles()
              << " " << 4*part.getNumTriangles() << "\n";

        PTriangleIterator triangleIt = part.triangleIterator();
        while (!triangleIt->end()) {
            PTriangle triangle = triangleIt->next();
            plint i0 = triangle->vertex(0)->tag(vertexIDtagging);
            plint i1 = triangle->vertex(1)->tag(vertexIDtagging);
            plint i2 = triangle->vertex(2)->tag(vertexIDtagging);
            ofile << "3 " << i0 << " " << i1 << " " << i2 << "\n";
        }
        ofile << "\n";

        ofile << "CELL_TYPES " << part.getNumTriangles() << "\n";

        for (plint i=0; i<part.getNumTriangles(); ++i) {
            ofile << "5\n";
        }
        ofile << "\n";

        ofile << "POINT_DATA " << part.getNumVertices() << "\n";
        
        for (plint property=0; property<mesh.numVertexProperties(); ++property) {
            ofile << "SCALARS " << mesh.getVertexPropertyName(property)
                  << (sizeof(T)==sizeof(double) ? " double" : " float")
                  << " 1\n"
                  << "LOOKUP_TABLE default\n";
            vertexIt = part.vertexIterator();
            while (!vertexIt->end()) {
                PVertex vertex = vertexIt->next();
                ofile << vertex->property(property) << "\n";
            }
            ofile << "\n";
        }
    }
}

}  // namespace plb


#endif  // VTK_FILE_IO_HH

