/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef X3D_FILE_IO_HH
#define X3D_FILE_IO_HH


#include <limits>
#include <fstream>
#include "geometry/x3dFileIO.h"
#include "io/colormaps.h"

namespace plb {


template<typename T>
void writeX3D_json(RawConnectedTriangleMesh<T> mesh, FileName fname, Array<T,3>& location, T& scale)
{
    if (!global::mpi().isMainProcessor()) return;
    if (mesh.getNumTriangles() == 0) return;
    typedef typename ConnectedTriangleMesh<T>::PTriangleIterator PTriangleIterator;
    typedef typename ConnectedTriangleMesh<T>::PVertexIterator PVertexIterator;
    typedef typename ConnectedTriangleMesh<T>::PTriangle PTriangle;
    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

    std::ofstream ofile(fname.get().c_str());

    if (scale<0.) {
        x3d_transform(mesh, location, scale);
    }

    to_x3d_units(mesh, location, scale);

    ofile << "{";
    ofile << "\"globalProperties\":{";
    ofile << "\"scale\":" << scale << ",\n";
    ofile << "\"offset\":["
          << location[0] << "," << location[1] << "," << location[2] << "],\n"
          << "\"units\":\"m\""
          << "},\n";

    ofile << "\"geometry\":[{";

    for (plint iPart=0; iPart<mesh.numParts(); ++iPart) {
        RawConnectedTriangleMesh<T> part = extractConnectedPart(mesh, iPart);

        plint vertexIDtagging = part.getVertexTag("UniqueID");
        std::map<plint,plint> toNewVertexID;
        PVertexIterator vertexIt = part.vertexIterator();
        plint newVertexID=0;
        while (!vertexIt->end()) {
            PVertex vertex(vertexIt->next());
            toNewVertexID[vertex->tag(vertexIDtagging)] = newVertexID;
            ++newVertexID;
        }

        std::string geometry_label = "Part_"+util::val2str(iPart);
        ofile << "\"part\":\"" << geometry_label << "\",\n";
        ofile << "\"title\":\"" << mesh.partName(iPart) << "\",\n";
        ofile << "\"description\":\"\",\n";
        ofile << "\"coordIndex\":[";
        PTriangleIterator triangleIt = part.triangleIterator();
        plint numTriangles=0;
        while (!triangleIt->end()) {
            PTriangle triangle = triangleIt->next();
            plint i0 = triangle->vertex(0)->tag(vertexIDtagging);
            plint i1 = triangle->vertex(1)->tag(vertexIDtagging);
            plint i2 = triangle->vertex(2)->tag(vertexIDtagging);
            if (numTriangles!=0) ofile << ",";
            ofile << toNewVertexID[i0] << "," << toNewVertexID[i1] << "," << toNewVertexID[i2];
            numTriangles++;
        }
        ofile << "],\n";
        ofile << "\"coordPoint\":[";
        vertexIt = part.vertexIterator();
        plint numVertices=0;
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            if (numVertices!=0) ofile << ",";
            ofile << (*vertex)[0] << "," << (*vertex)[1] << "," << (*vertex)[2];
            ++numVertices;
        }
        ofile << "],\n";
        ofile << "\"normalVector\":[";
        vertexIt = part.vertexIterator();
        numVertices=0;
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            Array<T,3> n = vertex->normal();
            if (numVertices!=0) ofile << ",";
            ofile << n[0] << "," << n[1] << "," << n[2];
            ++numVertices;
        }
        ofile << "]";
        if (iPart<mesh.numParts()-1) {
            ofile << "},\n{";
        }
    }
    ofile << "}],\n";

    ofile << "\"coloredProperties\":[{";
    plint numColorProp = mesh.numVertexProperties();
    for (plint iProp=0; iProp<numColorProp; ++iProp) {
        T minProp = std::numeric_limits<T>::max();
        T maxProp = std::numeric_limits<T>::min();
        PVertexIterator vertexIt = mesh.vertexIterator();
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            T prop = vertex->property(iProp);
            if (prop>maxProp) maxProp = prop;
            if (prop<minProp) minProp = prop;
        }
        ofile << "\"property\":\"" << mesh.getVertexPropertyName(iProp) << "\",\n";
        ofile << "\"title\":\"" << mesh.getVertexPropertyName(iProp) << "\",\n";
        ofile << "\"min\":" << minProp << ",\n";
        ofile << "\"max\":" << maxProp << ",\n";
        ofile << "\"units\": \"" << "NA" << "\",\n";
        ofile << "\"geometry\":[{";
        for (plint iPart=0; iPart<mesh.numParts(); ++iPart) {
            RawConnectedTriangleMesh<T> part = extractConnectedPart(mesh, iPart);
            std::string geometry_label = "Part_"+util::val2str(iPart);
            ofile << "\"part\":\"" << geometry_label << "\",\n"
                  << "\"vertexValues\":[";
            vertexIt = part.vertexIterator();
            plint iVertex=0;
            while (!vertexIt->end()) {
                PVertex vertex = vertexIt->next();
                T prop = vertex->property(iProp);
                if (iVertex!=0) ofile << ",";
                ofile << (prop-minProp)/(maxProp-minProp);
                ++iVertex;
            }
            ofile << "]";
            if (iPart<mesh.numParts()-1) {
                ofile << "},\n{";
            }
        }
        ofile << "}]";
        if (iProp<numColorProp-1) {
            ofile << "},\n{";
        }
    }
    ofile << "}]";
    ofile << "}";
}

template<typename T>
void x3d_transform( RawConnectedTriangleMesh<T> const& mesh,
                    Array<T,3>& location, T& scale )
{
    Cuboid<T> bound = mesh.getBoundingCuboid();
    location = (T)0.5*(bound.lowerLeftCorner+bound.upperRightCorner);
    Array<T,3> delta = bound.upperRightCorner - bound.lowerLeftCorner;
    T maxDelta = std::max(std::max(delta[0],delta[1]),delta[2]);
    scale = 0.5*maxDelta;
}

template<typename T>
void writeX3D_blender (
        RawConnectedTriangleMesh<T> mesh, FileName fname,
        Array<T,3>& location, T& scale, bool htmlFormat,
        std::string propertyName, std::string colormap )
{
    if (!global::mpi().isMainProcessor()) return;
    typedef typename ConnectedTriangleMesh<T>::PTriangleIterator PTriangleIterator;
    typedef typename ConnectedTriangleMesh<T>::PVertexIterator PVertexIterator;
    typedef typename ConnectedTriangleMesh<T>::PTriangle PTriangle;
    typedef typename ConnectedTriangleMesh<T>::PVertex PVertex;

    std::ofstream ofile(fname.get().c_str());

    if (scale<0.) {
        x3d_transform(mesh, location, scale);
    }

    to_x3d_units(mesh, location, scale);

    ofile << std::fixed << std::setprecision(5);
    if (htmlFormat) {
        ofile << "<!DOCTYPE HTML>\n";
        ofile << "<html>\n";
        ofile << "    <head>\n";
        ofile << "        <title>" << "Palabos Geometry" << "</title>\n";
        ofile << "        <link rel=\"stylesheet\" type=\"text/css\" href=\"http://www.x3dom.org/download/x3dom.css\">\n";
        ofile << "        <script type=\"text/javascript\" src=\"http://www.x3dom.org/download/1.4/x3dom-full.js\">\n";
        ofile << "        </script>\n";

        ofile << "        <script type=\"text/javascript\">\n";
        ofile << "            var scale = " << std::setprecision(5) << scale << ";\n";
        ofile << "            var offset = new Array(" << std::setprecision(5)
              << location[0] << ","
              << location[1] << ","
              << location[2] << ");\n";
        ofile << "        </script>\n";
        ofile << "    </head>\n";
        ofile << "    <X3D style=\"width: 100%; height :100%\">\n";
    }
    else {
        ofile << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
        ofile << "<!DOCTYPE X3D PUBLIC \"ISO//Web3D//DTD X3D 3.0//EN\" \"http://www.web3d.org/specifications/x3d-3.0.dtd\">\n";
        ofile << "<X3D version=\"3.0\" profile=\"Immersive\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema-instance"
              << "\" xsd:noNamespaceSchemaLocation=\"http://www.web3d.org/specifications/x3d-3.0.xsd\">\n";
    }
    ofile << "        <Scene>\n";
    ofile << "            <Shape>\n";
    // Note: in the following, we say that the colors and normals are per-vertex.
    // if instead we had set the values to "false", they would be per triangle.
    // Note that if, in the vertex case, the number of indicated vertex colors was
    // different than the total number of vertices (or idem in the case of triangles),
    // it would be mandatory to set normalIndex="..." or colorIndex="..."
    // to relate the color values or normals to their vertex or triangle.
    ofile << "            <IndexedFaceSet solid=\"false\" colorPerVertex=\"true\" "
          <<                              "normalPerVertex=\"true\" coordIndex=\"";

    plint vertexIDtagging = mesh.getVertexTag("UniqueID");
    std::map<plint,plint> toNewVertexID;
    PVertexIterator vertexIt = mesh.vertexIterator();
    plint newVertexID=0;
    while (!vertexIt->end()) {
        PVertex vertex(vertexIt->next());
        toNewVertexID[vertex->tag(vertexIDtagging)] = newVertexID;
        ++newVertexID;
    }

    PTriangleIterator triangleIt = mesh.triangleIterator();
    plint numTriangles=0;
    while (!triangleIt->end()) {
        PTriangle triangle = triangleIt->next();
        plint i0 = triangle->vertex(0)->tag(vertexIDtagging);
        plint i1 = triangle->vertex(1)->tag(vertexIDtagging);
        plint i2 = triangle->vertex(2)->tag(vertexIDtagging);
        ofile << toNewVertexID[i0] << " " << toNewVertexID[i1] << " " << toNewVertexID[i2] << " -1 ";
        numTriangles++;
    }
    ofile << "\">\n";
    ofile << "            <Coordinate point=\"";

    vertexIt = mesh.vertexIterator();
    while (!vertexIt->end()) {
        PVertex vertex = vertexIt->next();
        ofile << (*vertex)[0] << " " << (*vertex)[1] << " " << (*vertex)[2] << " ";
    }
    if (htmlFormat) {
        ofile << "\"> </Coordinate>\n";
    }
    else {
        ofile << "\"/>\n";
    }
    ofile << "            <Normal vector=\"";

    vertexIt = mesh.vertexIterator();
    while (!vertexIt->end()) {
        PVertex vertex = vertexIt->next();
        Array<T,3> n = vertex->normal();
        ofile << n[0] << " " << n[1] << " " << n[2] << " ";
    }
    if (htmlFormat) {
        ofile << "\"> </Normal>\n";
    }
    else {
        ofile << "\"/>\n";
    }

    plint iProp = mesh.getVertexProperty(propertyName);
    if(iProp!=-1) {
        ColorMap map = mapGenerators::generateMap(colormap);
        ofile << "            <Color color=\"";
        T minProp = std::numeric_limits<T>::max();
        T maxProp = std::numeric_limits<T>::min();
        PVertexIterator vertexIt = mesh.vertexIterator();
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            T prop = vertex->property(iProp);
            if (prop>maxProp) maxProp = prop;
            if (prop<minProp) minProp = prop;
        }
        vertexIt = mesh.vertexIterator();
        while (!vertexIt->end()) {
            PVertex vertex = vertexIt->next();
            T prop = vertex->property(iProp);
            rgb value = map.get((prop-minProp)/(maxProp-minProp));
            ofile << value.r << " " << value.g << " " << value.b << ", ";
        }
        if (htmlFormat) {
            ofile << "\"> </Color>\n";
        }
        else {
            ofile << "\"/>\n";
        }
    }
    ofile << "            </IndexedFaceSet>\n";
    ofile << "            </Shape>\n";
    ofile << "        </Scene>\n";
    ofile << "    </X3D>\n";
}

template<typename T>
void to_x3d_units(RawConnectedTriangleMesh<T>& mesh, Array<T,3> location, T scale) {
    mesh.translate(-location);
    mesh.scale(1./scale);
}

}  // namespace plb

#endif  // X3D_FILE_IO_HH


